__all__ = [
    "DoubleDuelingRDQN",
    "DoubleDuelingRDQNConfig",
    "evaluate",
    "train"
]

from l2rpn_baselines.DoubleDuelingRDQN.DoubleDuelingRDQN import DoubleDuelingRDQN
from l2rpn_baselines.DoubleDuelingRDQN.DoubleDuelingRDQNConfig import DoubleDuelingRDQNConfig
from l2rpn_baselines.DoubleDuelingRDQN.evaluate import evaluate
from l2rpn_baselines.DoubleDuelingRDQN.train import train




__all__ = [
    "DuelQSimple",
    "evaluate",
    "train",
    "DuelQ_NNParam"
]

from l2rpn_baselines.DuelQSimple.DuelQSimple import DuelQSimple
from l2rpn_baselines.DuelQSimple.evaluate import evaluate
from l2rpn_baselines.DuelQSimple.train import train
from l2rpn_baselines.DuelQSimple.DuelQ_NNParam import DuelQ_NNParam



from grid2op.Runner import Runner
from grid2op.Episode import EpisodeReplay
try:
    import os
    import shutil
except:
    print("You need the following package to train the A3C baseline")
    print(sys.exc_info())
    exit()

def run(runner_params_from_env,NB_EPISODE,agent_path,max_steps_in_episode,agent,gif_save,agent_name):
    # Build runner
    runner = Runner(**runner_params_from_env,
                    agentClass=None,
                    agentInstance=agent)

    PATH_SAVE = os.path.join(agent_path, agent_name)#agent_path + "\\" + agent_name
    PATH_SAVE2 = agent_path + r"\_cache"

    # Check if there are any previously stored results. If they are available then delete them and use the "Runner" to
    # evaluate more scenarios using the agent.
    if os.path.exists(PATH_SAVE):
        shutil.rmtree(PATH_SAVE)
    # WE ALSO REMOVE THE __CACHE WHICH IS CREATED BY THE GRID2VIZ. (This is useful especially when the agent's
    # performance data or the scenario data is overwritten. Since the grid2viz then needs the cache to be reset).
    # Here we always reset it to ensure the grid2viz results are displaying the most recently generated agent's behavior
    # correctly. (This will increase the running time of grid2viz to display initial results).
    if os.path.exists(PATH_SAVE2):
        shutil.rmtree(PATH_SAVE2)

    os.makedirs(agent_path, exist_ok=True)
    # NOTE: This code does not support the "runner.run" to use multiple cores.
    res = runner.run(nb_episode=NB_EPISODE, path_save=PATH_SAVE,max_iter=max_steps_in_episode)
    # print(agent.action_list)

    if gif_save:
        # Parameters to save the gif image of the performance of the agent.
        gif_name = "episode"
        ep_replay = EpisodeReplay(agent_path=PATH_SAVE)

    print("The results for the trained agent are:")
    for _, chron_name, cum_reward, nb_time_step, max_ts in res:
        msg_tmp = "\tFor chronics located at {}\n".format(chron_name)
        msg_tmp += "\t\t - cumulative reward: {:.6f}\n".format(cum_reward)
        msg_tmp += "\t\t - number of time steps completed: {:.0f} / {:.0f}".format(nb_time_step, max_ts)
        print(msg_tmp)

        if gif_save:
            # Uncomment the below code to save the gif image to "PATH_SAVE"
            ep_replay.replay_episode(chron_name,  # which chronic was started
                                 gif_name=gif_name, # Name of the gif file
                                 display=False,  # dont wait before rendering each frames
                                 fps=3.0)  # limit to 3 frames per second



#!/usr/bin/env python3

# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

from collections import deque
import random
import numpy as np


class ExperienceBuffer:

    def __init__(self, buffer_size, batch_size, trace_length):
        self.buffer_size = buffer_size
        self.batch_size = batch_size
        self.trace_length = trace_length
        self.buffer = [[]]
        self.current_episode = 0

    def add(self, s, a, r, d, s2, e):
        experience = np.reshape(np.array([s, a, r, d, s2]), [1,5])

        # New episode
        if self.current_episode < e:
            self.current_episode = e
            
            if self.size_episode() < self.trace_length:
                # Last episode was too short, reuse buffer
                self.buffer[-1] = []
            else:
                # Create new episode entry
                self.buffer.append([])
                # Forget oldest episode to respect size limits
                if len(self.buffer) > self.buffer_size:
                    self.buffer.pop(0)

        # Save experience
        self.buffer[-1].append(experience)

    def size(self):
        return len(self.buffer)

    def size_episode(self, episode_index = -1):
        return len(self.buffer[episode_index])

    def can_sample(self):
        if self.size() < self.batch_size:
            return False
        if self.size_episode() < self.trace_length:
            return False
        return True

    def sample(self):
        samples = []
        # Get random episodes
        sampled_episodes = random.sample(self.buffer, self.batch_size)
        # Get random trace in each episode
        for episode in sampled_episodes:
            max_trace_start = len(episode) - self.trace_length
            trace_start = np.random.randint(0, max_trace_start + 1)
            trace_end = trace_start + self.trace_length
            sample = episode[trace_start:trace_end]
            samples.append(sample)

        samples = np.array(samples)
        return np.reshape(samples, [self.batch_size * self.trace_length, 5])

    def clear(self):
        self.buffer = [[]]
        self.current_episode = 0


if __name__ == "__main__":
    REPLAY_BUFFER_SIZE = 128
    BATCH_SIZE = 32
    TRACE_LENGTH = 8
    rbuf = ExperienceBuffer(REPLAY_BUFFER_SIZE, BATCH_SIZE, TRACE_LENGTH)
    for ep_idx in range(BATCH_SIZE + 1):
        for t_idx in range(TRACE_LENGTH + 1): #range(np.random.randint(0, 32)):
            rbuf.add([0] * 434, "a", "r", False, [4] * 434, ep_idx)
            can_sample = rbuf.can_sample()
            buf_info = "ep_idx={}, t_idx={}, can_sample={}"
            print (buf_info.format(ep_idx, t_idx, str(can_sample)))
            if can_sample:
                batch = rbuf.sample()
                s2_batch = np.vstack(batch[:, 4])
                s2_batch = s2_batch.reshape(BATCH_SIZE, TRACE_LENGTH, 434)
                print (s2_batch.shape)



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.
__version__ = "0.4.1"

__all__ = [
    "cli_eval",
           "cli_train",
           "str2bool",
           "save_log_gif",
           "make_multi_env",
           "zip_for_codalab",
           "train_generic",
           "TrainingParam",
           "NNParam",
           "ReplayBuffer",
           "BaseDeepQ",
           "DeepQAgent"
]

from l2rpn_baselines.utils.cli_eval import cli_eval
from l2rpn_baselines.utils.cli_train import cli_train
from l2rpn_baselines.utils.str2bool import str2bool
from l2rpn_baselines.utils.save_log_gif import save_log_gif
from l2rpn_baselines.utils.zip_for_codalab import zip_for_codalab
from l2rpn_baselines.utils.train_generic import train_generic
from l2rpn_baselines.utils.make_multi_env import make_multi_env

from l2rpn_baselines.utils.TrainingParam import TrainingParam
from l2rpn_baselines.utils.NNParam import NNParam
from l2rpn_baselines.utils.ReplayBuffer import ReplayBuffer
from l2rpn_baselines.utils.BaseDeepQ import BaseDeepQ
from l2rpn_baselines.utils.DeepQAgent import DeepQAgent



#!/usr/bin/env python3

# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import os
from grid2op.Runner import Runner

from l2rpn_baselines.Template.Template import Template
from l2rpn_baselines.utils.save_log_gif import save_log_gif


def evaluate(env,
             load_path=".",
             logs_path=None,
             nb_episode=1,
             nb_process=1,
             max_steps=-1,
             verbose=False,
             save_gif=False,
             **kwargs):
    """
    In order to submit a valid basline, it is mandatory to provide a "evaluate" function with the same signature as this one.

    Parameters
    ----------
    env: :class:`grid2op.Environment.Environment`
        The environment on which the baseline will be evaluated.

    load_path: ``str``
        The path where the model is stored. This is used by the agent when calling "agent.load)

    logs_path: ``str``
        The path where the agents results will be stored.

    nb_episode: ``int``
        Number of episodes to run for the assessment of the performance.
        By default it's 1.

    nb_process: ``int``
        Number of process to be used for the assessment of the performance.
        Should be an integer greater than 1. By defaults it's 1.

    max_steps: ``int``
        Maximum number of timestep each episode can last. It should be a positive integer or -1.
        -1 means that the entire episode is run (until the chronics is out of data or until a game over).
        By default it's -1.

    verbose: ``bool``
        verbosity of the output

    save_gif: ``bool``
        Whether or not to save a gif into each episode folder corresponding to the representation of the said episode.

    kwargs:
        Other key words arguments that you are free to use for either building the agent save it etc.

    Returns
    -------
    ``None``
    """
    runner_params = env.get_params_for_runner()
    runner_params["verbose"] = verbose

    # Create the agent (this piece of code can change)
    agent = Template(env.action_space, env.observation_space, "Template")

    # Load weights from file (for example)
    agent.load(load_path)

    # Build runner
    runner = Runner(**runner_params,
                    agentClass=None,
                    agentInstance=agent)

    # you can do stuff with your model here

    # start the runner
    res = runner.run(path_save=logs_path,
                     nb_episode=nb_episode,
                     nb_process=nb_process,
                     max_iter=max_steps,
                     pbar=False)

    # Print summary
    print("Evaluation summary:")
    for _, chron_name, cum_reward, nb_time_step, max_ts in res:
        msg_tmp = "\tFor chronics located at {}\n".format(chron_name)
        msg_tmp += "\t\t - cumulative reward: {:.6f}\n".format(cum_reward)
        msg_tmp += "\t\t - number of time steps completed: {:.0f} / {:.0f}".format(nb_time_step, max_ts)
        print(msg_tmp)

    if save_gif:
        save_log_gif(logs_path, res)

if __name__ == "__main__":
    """
    This is a possible implementation of the eval script.
    """
    import grid2op
    from l2rpn_baselines.utils import cli_eval
    args_cli = cli_eval().parse_args()
    env = grid2op.make()
    evaluate(env,
             load_path=args_cli.load_path,
             logs_path=args_cli.logs_path,
             nb_episode=args_cli.nb_episode,
             nb_process=args_cli.nb_process,
             max_steps=args_cli.max_steps,
             verbose=args_cli.verbose,
             save_gif=args_cli.save_gif)



#!/usr/bin/env python3

# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.


import argparse
import json
import tensorflow as tf
import numpy as np

from grid2op.MakeEnv import make2
from grid2op.Action import PowerlineChangeAndDispatchAction

from l2rpn_baselines.DoubleDuelingDQN.DoubleDuelingDQN import DoubleDuelingDQN


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.bool_):
            return bool(obj)
        else:
            return super(NpEncoder, self).default(obj)


def cli():
    parser = argparse.ArgumentParser(description="Action space inspector")
    parser.add_argument("--path_data", required=True,
                        help="Path to the dataset root directory")
    return parser.parse_args()


def prune_impact_bool(impact_section, bool_key):
    if impact_section is None:
        return None
    if bool_key not in impact_section:
        return None
    if impact_section[bool_key] == False:
        return None

    return impact_section


def prune_impact_count(impact_section, count_key):
    if impact_section is None:
        return None
    if count_key not in impact_section:
        return None
    if impact_section[count_key] == 0:
        return None
    
    impact_section.pop(count_key)
    return impact_section


def prune_impact_array(impact_section, array_key):
    if impact_section is None:
        return None
    if array_key not in impact_section:
        return None
    if len(impact_section[array_key]) == 0:
        return None

    return impact_section[array_key]


def print_actions(agent):
    actions_dict = {}
    for i in range(agent.action_size):
        index_str = str(i)
        action = agent.convert_act(i)
        impact = action.impact_on_objects()
        pruned_impact = {
            "injection": prune_impact_bool(impact["injection"], "changed"),
            "line_reconnect": prune_impact_count(impact["force_line"]["reconnections"], "count"),
            "line_disconnect": prune_impact_count(impact["force_line"]["disconnections"], "count"),
            "connect_bus": prune_impact_array(impact["topology"], "assigned_bus"),
            "disconnect_bus": prune_impact_array(impact["topology"], "disconnect_bus"),
            "switch_line": prune_impact_array(impact["switch_line"], "powerlines"),
            "switch_bus": prune_impact_array(impact["topology"], "bus_switch"),
            "redispatch": prune_impact_array(impact["redispatch"], "generators")
        }
        compact_impact = {k: v for k, v in pruned_impact.items() if v is not None}
        actions_dict[index_str] = compact_impact

    actions_json = json.dumps(actions_dict,
                              indent=2,
                              cls=NpEncoder)
    print (actions_json)
        

if __name__ == "__main__":
    args = cli()
    env = make2(args.path_data, action_class=PowerlineChangeAndDispatchAction)
    # Limit gpu usage
    physical_devices = tf.config.list_physical_devices('GPU')
    tf.config.experimental.set_memory_growth(physical_devices[0], True)

    agent = DoubleDuelingDQN(env,
                             env.action_space,
                             is_training=False)
    print_actions(agent)



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import numpy as np

# tf2.0 friendly
import warnings

import tensorflow as tf
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=FutureWarning)
    from tensorflow.keras.models import Sequential, Model
    from tensorflow.keras.layers import Activation
    from tensorflow.keras.layers import Input, Lambda, subtract, add
    import tensorflow.keras.backend as K

from l2rpn_baselines.utils import BaseDeepQ, TrainingParam

# try:
#     from leap_net import Ltau  # this import might change if you use the "quick and dirty way".
# except ImportError:
#     # Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
#     # See AUTHORS.txt
#     # This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
#     # If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
#     # you can obtain one at http://mozilla.org/MPL/2.0/.
#     # SPDX-License-Identifier: MPL-2.0
#     # This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.
# MSG_WARNING = "Leap net model is not installed on your system. Please visit \n" \
#               "https://github.com/BDonnot/leap_net \n" \
#               "to have the latest Leap net implementation."
# warnings.warn(MSG_WARNING)

# TODO implement that in the leap net package too
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import add as tfk_add
from tensorflow.keras.layers import multiply as tfk_multiply


class LtauBis(Layer):
    """
    This layer implements the Ltau layer.

    This kind of leap net layer computes, from their input `x`: `d.(e.x * tau)` where `.` denotes the
    matrix multiplication and `*` the elementwise multiplication.
    """

    def __init__(self, initializer='glorot_uniform', use_bias=True, trainable=True, name=None, **kwargs):
        super(LtauBis, self).__init__(trainable=trainable, name=name, **kwargs)
        self.initializer = initializer
        self.use_bias = use_bias
        self.e = None
        self.d = None

    def build(self, input_shape):
        is_x, is_tau = input_shape
        nm_e = None
        nm_d = None
        if self.name is not None:
            nm_e = '{}_e'.format(self.name)
            nm_d = '{}_d'.format(self.name)
        self.e = Dense(is_tau[-1],
                       kernel_initializer=self.initializer,
                       use_bias=self.use_bias,
                       trainable=self.trainable,
                       name=nm_e)
        self.d = Dense(is_x[-1],
                       kernel_initializer=self.initializer,
                       use_bias=False,
                       trainable=self.trainable,
                       name=nm_d)

    def get_config(self):
        config = super().get_config().copy()
        config.update({
            'initializer': self.initializer,
            'use_bias': self.use_bias
        })
        return config

    def call(self, inputs, **kwargs):
        x, tau = inputs
        tmp = self.e(x)
        tmp = tfk_multiply([tau, tmp])  # element wise multiplication
        res = self.d(tmp)  # no addition of x
        # res = tfk_add([x, tmp])
        return res


class DuelQLeapNet_NN(BaseDeepQ):
    """
    Constructs the desired duelling deep q learning network with a leap neural network as a modeling
    of the q function
    """
    def __init__(self,
                 nn_params,
                 training_param=None):
        if training_param is None:
            training_param = TrainingParam()
        BaseDeepQ.__init__(self,
                           nn_params,
                           training_param)
        self._custom_objects = {"LtauBis": LtauBis}
        self.construct_q_network()
        self._max_global_norm_grad = training_param.max_global_norm_grad
        self._max_value_grad = training_param.max_value_grad
        self._max_loss = training_param.max_loss

    def construct_q_network(self):
        """
        First the :attr:`l2rpn_baselines.BaseDeepQ.nn_archi` parameters are used to create a neural network
        to 'encode' the data. Then the leaps occur.

        Afterward the model is split into value an advantage, and treated as usually in any D3QN.

        """
        # Uses the network architecture found in DeepMind paper
        # The inputs and outputs size have changed, as well as replacing the convolution by dense layers.
        self._model = Sequential()
        input_x = Input(shape=(self._nn_archi.x_dim,),
                        name="x")
        inputs_tau = [Input(shape=(el,), name="tau_{}".format(nm_)) for el, nm_ in
                      zip(self._nn_archi.tau_dims, self._nn_archi.list_attr_obs_tau)]

        lay = input_x
        for (size, act) in zip(self._nn_archi.sizes, self._nn_archi.activs):
            lay = Dense(size)(lay)  # put at self.action_size
            lay = Activation(act)(lay)

        # TODO multiple taus
        l_tau = lay
        for el, nm_ in zip(inputs_tau, self._nn_archi.list_attr_obs_tau):
            l_tau = l_tau + LtauBis(name="leap_{}".format(nm_))([lay, el])

        advantage = Dense(self._action_size)(l_tau)
        value = Dense(1, name="value")(l_tau)

        meaner = Lambda(lambda x: K.mean(x, axis=1))
        mn_ = meaner(advantage)
        tmp = subtract([advantage, mn_])
        policy = add([tmp, value], name="policy")

        self._model = Model(inputs=[input_x, *inputs_tau], outputs=[policy])
        self._schedule_model, self._optimizer_model = self.make_optimiser()
        self._model.compile(loss='mse', optimizer=self._optimizer_model)

        self._target_model = Model(inputs=[input_x, *inputs_tau], outputs=[policy])

    def _make_x_tau(self, data):
        data_x = data[:, :self._nn_archi.x_dim]

        # for the taus
        data_tau = []
        prev = self._nn_archi.x_dim
        for sz, add_, mul_ in zip(self._nn_archi.tau_dims, self._nn_archi.tau_adds, self._nn_archi.tau_mults):
            data_tau.append((data[:, prev:prev+sz] + add_) * mul_)
            prev += sz

        res = [data_x, *data_tau]
        return res

    def predict_movement(self, data, epsilon, batch_size=None):
        """Predict movement of game controler where is epsilon
        probability randomly move."""
        if batch_size is None:
            batch_size = data.shape[0]
        data_split = self._make_x_tau(data)
        res = super().predict_movement(data_split, epsilon=epsilon, batch_size=batch_size)
        return res

    def train(self, s_batch, a_batch, r_batch, d_batch, s2_batch, tf_writer=None, batch_size=None):
        if batch_size is None:
            batch_size = s_batch.shape[0]
        s_batch_split = self._make_x_tau(s_batch)
        s2_batch_split = self._make_x_tau(s2_batch)
        res = super().train(s_batch_split,
                            a_batch,
                            r_batch,
                            d_batch,
                            s2_batch_split,
                            tf_writer=tf_writer,
                            batch_size=batch_size)
        return res

    def train_on_batch(self, model, optimizer_model, x, y_true):
        """
        clip the loss
        """
        with tf.GradientTape() as tape:
            # Get y_pred for batch
            y_pred = model(x)
            # Compute loss for each sample in the batch
            # and then clip it
            batch_loss = self._clipped_batch_loss(y_true, y_pred)
            # Compute mean scalar loss
            loss = tf.math.reduce_mean(batch_loss)
        loss_npy = loss.numpy()

        # Compute gradients
        grads = tape.gradient(loss, model.trainable_variables)

        # clip gradients
        if self._max_global_norm_grad is not None:
            grads, _ = tf.clip_by_global_norm(grads, self._max_global_norm_grad)
        if self._max_value_grad is not None:
            grads = [tf.clip_by_value(grad, -self._max_value_grad, self._max_value_grad) for grad in grads]

        # Apply gradients
        optimizer_model.apply_gradients(zip(grads, model.trainable_variables))
        # Store LR
        self.train_lr = optimizer_model._decayed_lr('float32').numpy()
        # Return loss scalar
        return loss_npy

    def _clipped_batch_loss(self, y_true, y_pred):
        sq_error = tf.math.square(y_true - y_pred, name="sq_error")
        batch_sq_error = tf.math.reduce_sum(sq_error, axis=1, name="batch_sq_error")
        if self._max_loss is not None:
            res = tf.clip_by_value(batch_sq_error, 0.0, self._max_loss, name="batch_sq_error_clip")
        else:
            res = batch_sq_error
        return res


__all__ = [
    "SAC",
    "evaluate",
    "train",
    "SAC_NNParam"
]

from l2rpn_baselines.SAC.SAC import SAC
from l2rpn_baselines.SAC.evaluate import evaluate
from l2rpn_baselines.SAC.train import train
from l2rpn_baselines.SAC.SAC_NNParam import SAC_NNParam



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.
import os
import json
import numpy as np


class TrainingParam(object):
    """
    A class to store the training parameters of the models. It was hard coded in the getting_started/notebook 3
    of grid2op and put in this repository instead.

    Attributes
    ----------
    buffer_size: ``int``
        Size of the replay buffer
    minibatch_size: ``int``
        Size of the training minibatch
    update_freq: ``int``
        Frequency at which the model is trained. Model is trained once every `update_freq` steps using `minibatch_size`
        from an experience replay buffer.

    final_epsilon: ``float``
        value for the final epsilon (for the e-greedy)
    initial_epsilon: ``float``
        value for the initial epsilon (for the e-greedy)
    step_for_final_epsilon: ``int``
        number of step at which the final epsilon (for the epsilon greedy exploration) will be reached

    min_observation: ``int``
        number of observations before starting to train the neural nets. Before this number of iterations, the agent
        will simply interact with the environment.

    lr: ``float``
        The initial learning rate

    lr_decay_steps: ``int``
        The learning rate decay step

    lr_decay_rate: ``float``
        The learning rate decay rate

    num_frames: ``int``
        Currently not used

    discount_factor: ``float``
        The discount factor (a high discount factor is in favor of longer episode, a small one not really). This is
        often called "gamma" in some RL paper. It's the gamma in: "RL wants to minize the sum of the dicounted reward,
        which are sum_{t >= t_0} \gamma^{t - t_0} r_t

    tau: ``float``
        Update the target model. Target model is updated according to
        $target_model_weights[i] = self.training_param.tau * model_weights[i] + (1 - self.training_param.tau) * \
                                              target_model_weights[i]$

    min_iter: ``int``
        It is possible in the training schedule to limit the number of time steps an episode can last. This is mainly
        useful at beginning of training, to not get in a state where the grid has been modified so much the agent
        will never get into a state resembling this one ever again). Stopping the episode before this happens can
        help the learning.

    max_iter: ``int``
        Just like "min_iter" but instead of being the minimum number of iteration, it's the maximum.

    update_nb_iter: ``int``
        If max_iter_fun is the default one, this numer give the number of time we need to succeed a scenario before
        having to increase the maximum number of timestep allowed

    step_increase_nb_iter: ``int`` or  ``None``
        Of how many timestep we increase the maximum number of timesteps allowed per episode. Set it to O to deactivate
        this.

    max_iter_fun: ``function``
        A function that return the maximum number of steps an episode can count as for the current epoch. For example
        it can be `max_iter_fun = lambda epoch_num : np.sqrt(50 * epoch_num)`
        [default lambda x: x / self.update_nb_iter]

    oversampling_rate: ``float`` or ``None``
        Set it to None to deactivate the oversampling of hard scenarios. Otherwise, this oversampling is done
        with something like `proba = 1. / (time_step_lived**oversampling_rate + 1)` where `proba` is the probability
        to be selected at the next call to "reset" and `time_step_lived` is the number of time steps

    random_sample_datetime_start: ``int`` or ``None``
        If ``None`` during training the chronics will always start at the datetime the chronics start.
        Otherwise, the training scheme will skip a number of time steps between 0 and  `random_sample_datetime_start`
        when loading the next chronics. This is particularly useful when you want your agent to learn to operate
        the grid regardless of the hour of day or day of the week.

    update_tensorboard_freq: ``int``
        Frequency at which tensorboard is refresh (tensorboard summaries are saved every update_tensorboard_freq
        steps)

    save_model_each: ``int``
        Frequency at which the model is saved (it is saved every "save_model_each" steps)

    max_global_norm_grad: ``float``
        Maximum gradient norm allowed (can make the training more stable) default to None if deactivated.
        Not all baselines are compatible.

    max_value_grad: ``float``
        Maximum value the gradient can take. Assign it to ``None`` to deactivate it. This can make the training
        more stable in some cases, but can slow down the training process too. Not all baselines are compatible.

    max_loss: ``float``
        Clip the value of the loss function. Set it to ``None`` to deactivate it. Again, this can make the training
        more stable but possibly slower. Not all baselines are compatible.
    """
    _tol_float_equal = float(1e-8)

    _int_attr = ["buffer_size", "minibatch_size", "step_for_final_epsilon",
                  "min_observation", "last_step", "num_frames", "update_freq",
                 "min_iter", "max_iter", "update_tensorboard_freq", "save_model_each", "_update_nb_iter",
                 "step_increase_nb_iter"]
    _float_attr = ["_final_epsilon", "_initial_epsilon", "lr", "lr_decay_steps", "lr_decay_rate",
                    "discount_factor", "tau", "oversampling_rate",
                   "max_global_norm_grad", "max_value_grad", "max_loss"]

    def __init__(self,
                 buffer_size=40000,
                 minibatch_size=64,
                 step_for_final_epsilon=100000,  # step at which min_espilon is obtain
                 min_observation=5000,  # 5000
                 final_epsilon=1./(7*288.),  # have on average 1 random action per week of approx 7*288 time steps
                 initial_epsilon=0.4,
                 lr=1e-4,
                 lr_decay_steps=10000,
                 lr_decay_rate=0.999,
                 num_frames=1,
                 discount_factor=0.99,
                 tau=0.01,
                 update_freq=256,
                 min_iter=50,
                 max_iter=8064,  # 1 month
                 update_nb_iter=10,
                 step_increase_nb_iter=0,  # by default no oversampling / under sampling based on difficulty
                 update_tensorboard_freq=1000,  # update tensorboard every "update_tensorboard_freq" steps
                 save_model_each=10000,  # save the model every "update_tensorboard_freq" steps
                 random_sample_datetime_start=None,
                 oversampling_rate=None,
                 max_global_norm_grad=None,
                 max_value_grad=None,
                 max_loss=None
                 ):

        self.random_sample_datetime_start = random_sample_datetime_start

        self.buffer_size = buffer_size
        self.minibatch_size = minibatch_size
        self.min_observation = min_observation
        self._final_epsilon = float(final_epsilon)  # have on average 1 random action per day of approx 288 timesteps at the end (never kill completely the exploration)
        self._initial_epsilon = float(initial_epsilon)
        self.step_for_final_epsilon = float(step_for_final_epsilon)
        self.lr = lr
        self.lr_decay_steps = float(lr_decay_steps)
        self.lr_decay_rate = float(lr_decay_rate)

        # gradient clipping (if supported)
        self.max_global_norm_grad = max_global_norm_grad
        self.max_value_grad = max_value_grad
        self.max_loss = max_loss

        self.last_step = 0
        self.num_frames = int(num_frames)
        self.discount_factor = float(discount_factor)
        self.tau = float(tau)
        self.update_freq = update_freq
        self.min_iter = min_iter
        self.max_iter = max_iter
        self._update_nb_iter = update_nb_iter
        if step_increase_nb_iter is None:
            # 0 and None have the same effect: it disable the feature
            step_increase_nb_iter = 0
        self.step_increase_nb_iter = step_increase_nb_iter

        if oversampling_rate is not None:
            self.oversampling_rate = float(oversampling_rate)
        else:
            self.oversampling_rate = None

        self.update_tensorboard_freq = update_tensorboard_freq
        self.save_model_each = save_model_each
        self.max_iter_fun = self.default_max_iter_fun
        self._compute_exp_facto()

    @property
    def final_epsilon(self):
        return self._final_epsilon

    @final_epsilon.setter
    def final_epsilon(self, final_epsilon):
        self._final_epsilon = final_epsilon
        self._compute_exp_facto()

    @property
    def initial_epsilon(self):
        return self._initial_epsilon

    @initial_epsilon.setter
    def initial_epsilon(self, initial_epsilon):
        self._initial_epsilon = initial_epsilon
        self._compute_exp_facto()

    @property
    def update_nb_iter(self):
        return self._update_nb_iter

    @update_nb_iter.setter
    def update_nb_iter(self, update_nb_iter):
        self._update_nb_iter = update_nb_iter
        if self._update_nb_iter is not None and self._update_nb_iter > 0:
            self._1_update_nb_iter = 1.0 / self._update_nb_iter
        else:
            self._1_update_nb_iter = 1.0

    def _compute_exp_facto(self):
        if self.final_epsilon is not None and self.initial_epsilon is not None and self.final_epsilon > 0:
            self._exp_facto = np.log(self.initial_epsilon/self.final_epsilon)
        else:
            # TODO
            self._exp_facto = 1

    def default_max_iter_fun(self, nb_success):
        """the default max iteration function used"""
        return self.step_increase_nb_iter * int(nb_success * self._1_update_nb_iter)

    def tell_step(self, current_step):
        """tell this instance the number of training steps that have been made"""
        self.last_step = current_step

    def get_next_epsilon(self, current_step):
        """get the next epsilon for the e greedy exploration"""
        self.tell_step(current_step)
        if self.step_for_final_epsilon is None or self.initial_epsilon is None \
                or self._exp_facto is None or self.final_epsilon is None:
            res = 0.
        else:
            if current_step > self.step_for_final_epsilon:
                res = self.final_epsilon
            else:
                # exponential decrease
                res = self.initial_epsilon * np.exp(- (current_step / self.step_for_final_epsilon) * self._exp_facto )
        return res

    def to_dict(self):
        """serialize this instance to a dictionnary."""
        res = {}
        for attr_nm in self._int_attr:
            tmp = getattr(self, attr_nm)
            if tmp is not None:
                res[attr_nm] = int(tmp)
            else:
                res[attr_nm] = None
        for attr_nm in self._float_attr:
            tmp = getattr(self, attr_nm)
            if tmp is not None:
                res[attr_nm] = float(tmp)
            else:
                res[attr_nm] = None
        return res

    @staticmethod
    def from_dict(tmp):
        """initialize this instance from a dictionnary"""
        if not isinstance(tmp, dict):
            raise RuntimeError("TrainingParam from dict must be called with a dictionnary, and not {}".format(tmp))
        res = TrainingParam()
        for attr_nm in TrainingParam._int_attr:
            if attr_nm in tmp:
                tmp_ = tmp[attr_nm]
                if tmp_ is not None:
                    setattr(res, attr_nm, int(tmp_))
                else:
                    setattr(res, attr_nm, None)

        for attr_nm in TrainingParam._float_attr:
            if attr_nm in tmp:
                tmp_ = tmp[attr_nm]
                if tmp_ is not None:
                    setattr(res, attr_nm, float(tmp_))
                else:
                    setattr(res, attr_nm, None)

        res._compute_exp_facto()
        return res

    @staticmethod
    def from_json(json_path):
        """initialize this instance from a json"""
        if not os.path.exists(json_path):
            raise FileNotFoundError("No path are located at \"{}\"".format(json_path))
        with open(json_path, "r") as f:
            dict_ = json.load(f)
        return TrainingParam.from_dict(dict_)

    def save_as_json(self, path, name=None):
        """save this instance as a json"""
        res = self.to_dict()
        if name is None:
            name = "training_parameters.json"
        if not os.path.exists(path):
            raise RuntimeError("Directory \"{}\" not found to save the training parameters".format(path))
        if not os.path.isdir(path):
            raise NotADirectoryError("\"{}\" should be a directory".format(path))
        path_out = os.path.join(path, name)
        with open(path_out, "w", encoding="utf-8") as f:
            json.dump(res, fp=f, indent=4, sort_keys=True)

    def do_train(self):
        """return whether or not i should train the model at this time step"""
        return self.last_step % self.update_freq == 0

    def __eq__(self, other):
        res = True
        for el in self._int_attr:
            me_ = getattr(self, el)
            oth_ = getattr(other, el)
            if me_ != oth_:
                res = False
                break
        if res:
            for el in self._float_attr:
                me_ = getattr(self, el)
                oth_ = getattr(other, el)
                if me_ is None and oth_ is not None:
                    res = False
                    break
                if oth_ is None and me_ is not None:
                    res = False
                    break
                if me_ is None and oth_ is None:
                    continue
                if abs(float(me_) != float(oth_)) > self._tol_float_equal:
                    res = False
                    break
        return res



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.


def train_generic(agent,
                   env,
                   name="Template",
                   iterations=1,
                   save_path=None,
                   load_path=None,
                  **kwargs_train):
    """
    This function is a helper to train more easily some agent using their default "train" method.

    Parameters
    ----------
    agent: :class:`grid2op.Agent`
        A grid2op agent that must implement all the baseline attributes and the train method.

    env: :class:`grid2op.Environment`
        The environment on which to train your baseline. It must be compatible with the agent created.

    name: ``str``
        Here for compatibility with the baseline "train" method. Currently unused (define the name when you create
        your baseline)

    iterations: ``int``
        Number of iterations on which to train your agent.

    save_path: ``str``
        Where to save your results (put None do deactivate saving)

    load_path: ``str``
        Path to load the agent from.

    kwargs_train: ``dict``
        Other argument that will be passed to `agent.train(...)`

    Returns
    -------
    agent: :class:`grid2op.Agent`
        The trained agent.

    """

    if load_path is not None:
        agent.load(load_path)

    agent.train(env,
                iterations,
                save_path,
                **kwargs_train)

    return agent


if __name__ == "__main__":
    pass



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

from l2rpn_baselines.utils import DeepQAgent
DEFAULT_NAME = "DeepQSimple"


class DeepQSimple(DeepQAgent):
    """
    A simple deep q learning algorithm. It does nothing different thant its base class.
    """
    pass



# Copyright (C) 2018 - 2020 GEIRI North America, JEPC
# Authors: jiajaun <jiajun.duan@geirina.net>

import os
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import numpy as np


class DeepQNetworkDueling(object):
  def __init__(self,
               n_state=420,
               n_actions=175,
               learning_rate=1e-5,
               scope='dqn',
               summaries_dir=None):
    """
    Deep Q Network with experience replay and fixed Q-target

    Args:
    
      learning_rate: model learning rate
      scope: graph scope name
      summaries_dir: log directory of Tensorboard Filewriter

    Returns:
      pass
    """


    self.learning_rate = learning_rate
    self.n_state = n_state
    self.n_actions = n_actions
    self.scope = scope
    self.summary_writer = None

    # Build model and create Tensorboard summary
    with tf.variable_scope(scope):
      self._build_model()
      if summaries_dir:
        summary_dir = os.path.join(summaries_dir, "summaries_{}".format(scope))
        if not os.path.exists(summary_dir):
          os.makedirs(summary_dir)
        self.summary_writer = tf.summary.FileWriter(summary_dir)

  def _build_model(self):
    """
    Build Tensorflow graph
    """
    with tf.variable_scope('inputs'):
      # State as Vector
      self.s = tf.placeholder(
          shape=[None, self.n_state], dtype=tf.float32, name='s')
      # Placeholder for TD target
      self.q_target = tf.placeholder(
          shape=[None], dtype=tf.float32, name='q_target')
      # Placeholder for actions
      self.action_ph = tf.placeholder(
          shape=[None, 1], dtype=tf.int32, name='action')
      # Placeholder for Importance-Sampling(IS) weights
      self.IS_weights = tf.placeholder(
          shape=[None], dtype=tf.float32, name='IS_weights')

    # forward pass
    def forward_pass(s):
      
      s = tf.layers.batch_normalization(s, name = 'bn')
      fc1 = tf.layers.Dense(units=256, activation=tf.nn.relu, name='fc1')(s)

      # fc_V
      fc_V = tf.layers.Dense(64, activation=tf.nn.relu, name='fc_V_1')(fc1)
      self.fc_V = tf.layers.Dense(1, activation=None, name='fc_V_2')(fc_V)

      # fc_A
      fc_A = tf.layers.Dense(32, activation=tf.nn.relu, name='fc_A_1')(fc1)
      self.fc_A = tf.layers.Dense(self.n_actions, activation=None, name='fc_A_2')(fc_A)

      with tf.variable_scope('q_predict'):
        mean_A = tf.reduce_mean(self.fc_A, axis=1, keep_dims=True)
        q_predict = tf.add(self.fc_V, tf.subtract(self.fc_A, mean_A))

      return q_predict

    # forward pass
    self.q_predict = forward_pass(self.s)

    # q_predict for chosen action
    indexes = tf.reshape(tf.range(tf.shape(self.q_predict)[0]), shape=[-1, 1])
    action_indexes = tf.concat([indexes, self.action_ph], axis=1)
    self.q_predict_action = tf.gather_nd(self.q_predict, action_indexes)

    # calculate the mean batch loss
    self.abs_TD_errors = tf.abs(self.q_target - self.q_predict_action)
    self.loss_batch = self.IS_weights * \
        tf.squared_difference(self.q_target, self.q_predict_action)

    with tf.variable_scope('loss'):
      self.loss = tf.reduce_mean(self.loss_batch)

    # optimizer
    self.optimizer = tf.train.AdamOptimizer(learning_rate=self.learning_rate)

    # train_op
   # update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
    with tf.variable_scope('train_op'):
     # with tf.control_dependencies(update_ops):
      self.train_op = self.optimizer.minimize(
          self.loss, tf.train.get_global_step())

    # trainable parameters
    self.params_train = tf.get_collection(
        tf.GraphKeys.TRAINABLE_VARIABLES, scope=self.scope)

    # Summaries for Tensorboard
    # self.summaries = tf.summary.merge([
    #     tf.summary.histogram("loss_hist", self.loss_batch),
    #     tf.summary.histogram("q_values_hist", self.q_predict),
    # ])

  def epsilon_greedy(self, q_predict, epsilon):
    roll = np.random.uniform()
    if roll < epsilon:
      return np.random.randint(self.n_actions)
    else:
      return np.argmax(q_predict[0])

  def act(self, sess, s, epsilon):
    q_predict = sess.run(self.q_predict, feed_dict={self.s: s})
    return self.epsilon_greedy(q_predict, epsilon), q_predict[0]

  def predict(self, sess, s):
    return sess.run([self.q_predict], feed_dict={self.s: s})

  def update(self, sess, s, q_target, action, IS_weights):
    feed_dict = {self.s: s, self.q_target: q_target,
                 self.action_ph: action, self.IS_weights: IS_weights}
    _, loss, abs_TD_errors = sess.run(
        [self.train_op, self.loss, self.abs_TD_errors], feed_dict)
   # if self.summary_writer:
   #   self.summary_writer.add_summary(summaries)
    return loss, abs_TD_errors



# Original author Abhinav Sagar in https://github.com/abhinavsagar/Reinforcement-Learning-Tutorial
# Since then deleted

# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

from collections import deque
import numpy as np
import random
import pdb


class ReplayBuffer:
    """Constructs a buffer object that stores the past moves
    and samples a set of subsamples"""

    def __init__(self, buffer_size):
        self.buffer_size = buffer_size
        self.count = 0
        self.buffer = deque()

    def add(self, s, a, r, d, s2):
        """Add an experience to the buffer"""
        # S represents current state, a is action,
        # r is reward, d is whether it is the end,
        # and s2 is next state
        if np.any(~np.isfinite(s)) or np.any(~np.isfinite(s2)) or np.any(~np.isfinite(r)):
            # TODO proper handling of infinite values somewhere !!!!
            raise RuntimeError("Infinite value somwhere in at least one of the state")

        experience = (s, a, r, d, s2)
        if self.count < self.buffer_size:
            self.buffer.append(experience)
            self.count += 1
        else:
            self.buffer.popleft()
            self.buffer.append(experience)

    def size(self):
        return self.count

    def sample(self, batch_size):
        """Samples a total of elements equal to batch_size from buffer
        if buffer contains enough elements. Otherwise return all elements"""

        batch = []
        if self.count < batch_size:
            batch = random.sample(self.buffer, self.count)
        else:
            batch = random.sample(self.buffer, batch_size)

        # Maps each experience in batch in batches of states, actions, rewards
        # and new states
        s_batch, a_batch, r_batch, d_batch, s2_batch = list(map(np.array, list(zip(*batch))))
        return s_batch, a_batch, r_batch, d_batch, s2_batch

    def clear(self):
        self.buffer.clear()
        self.count = 0


try:
    import grid2op
    import threading
    import numpy as np
    import time
    import json
    import copy
    import os
    from grid2op import make
    from grid2op.Agent import MLAgent
    from grid2op.Environment import Environment
    from grid2op.Parameters import Parameters
    from grid2op.Reward import L2RPNReward, CombinedReward, CloseToOverflowReward, GameplayReward

    import tensorflow.compat.v1 as tf
    tf.disable_v2_behavior()

    from tensorflow.keras.layers import Dense, Input
    from tensorflow.keras.models import Model
    from tensorflow.keras.optimizers import Adam
    import tensorflow.python.keras.backend as K
except:
    print("You need the following package to train the A3C baseline")
    print(sys.exc_info())
    exit()

import l2rpn_baselines.Multithreading_agent.user_environment_make
# import user_environment_make

# Create the Agent instance here that can used with the Runner to test the performance of the trained RL agent.
class A3CAgent(MLAgent):
    # first change: An Agent must derived from grid2op.Agent (in this case MLAgent, because we manipulate vector instead
    # of classes) We will use this template to create our desired ML agent with unique neural network configuration.
    def __init__(self, state_size, action_size, env_name, action_space, value_multiplier,action_space_lists,
                 profiles_chronics,EPISODES_train2,time_step_end2,Hyperparameters,Thread_count,train_flag,save_path):
        MLAgent.__init__(self, action_space)
        # Parameter settings.
        # NOTE: MAKE SURE THE FOLLOWING SETTINGS ARE SAME AS THE TRAINED AGENT OR THE WEIGHTS WONT LOAD SUCCESSFULLY.
        # get size of state and action
        self.state_size = state_size
        self.action_size = action_size
        # get gym environment name
        self.env_name = env_name
        self.denominator = value_multiplier

        if train_flag:
            # these are hyper parameters for the A3C
            self.actor_lr = Hyperparameters["actor_learning_rate"]
            self.critic_lr = Hyperparameters["critic_learning_rate"]
            self.discount_factor = Hyperparameters["discount_factor"]
            self.threads = Thread_count
            self.save_path = os.path.abspath(save_path)
            if not os.path.exists(self.save_path):
                os.mkdir(self.save_path)
        else: #evaluating
            self.action_list = []

        self.hidden1, self.hidden2 = Hyperparameters["size_of_hidden_layer_1"], Hyperparameters["size_of_hidden_layer_2"]

        # create model for actor and critic network
        self.actor, self.critic = self.build_model()

        if train_flag:
            # method for training actor and critic network
            self.optimizer = [self.actor_optimizer(), self.critic_optimizer()]

            # global variables for threading
            global scores
            scores = []
            global time_step_end
            time_step_end = time_step_end2
            global EPISODES_train
            EPISODES_train = EPISODES_train2

            self.profiles_chronics = profiles_chronics

        # tf will use CPU. Number of GPU devices allowed to access are zero.
        self.sess = tf.InteractiveSession(config=tf.ConfigProto(device_count={'GPU': 0}))
        K.set_session(self.sess)
        self.sess.run(tf.global_variables_initializer())

        self.action_space_lists = action_space_lists
        self.gen_action_list = action_space_lists[0]
        self.load_action_list = action_space_lists[1]
        self.line_or_action_list = action_space_lists[2]
        self.line_ex_action_list = action_space_lists[3]

    # approximate policy and value using Neural Network
    # actor -> state is input and probability of each action is output of network
    # critic -> state is input and value of state is output of network
    # actor and critic network share first hidden layer. These neural networks should be same as the neural network
    # from "train.py".
    def build_model(self):
        state = Input(batch_shape=(None,  self.state_size))
        shared = Dense(self.hidden1, input_dim=self.state_size, activation='relu', kernel_initializer='glorot_uniform')(state)

        actor_hidden = Dense(self.hidden2, activation='relu', kernel_initializer='glorot_uniform')(shared)
        action_prob = Dense(self.action_size, activation='softmax', kernel_initializer='glorot_uniform')(actor_hidden)
        # action_prob = K.softmax(action_intermediate)

        value_hidden = Dense(self.hidden2, activation='relu', kernel_initializer='he_uniform')(shared)
        state_value = Dense(1, activation='linear', kernel_initializer='he_uniform')(value_hidden)

        actor = Model(inputs=state, outputs=action_prob)
        critic = Model(inputs=state, outputs=state_value)

        actor._make_predict_function()
        critic._make_predict_function()

        # actor.summary()
        # critic.summary()

        return actor, critic

    def act(self, state_as_dict, reward, done=False):
        state = useful_state(state_as_dict,self.denominator)
        state = state.reshape([1,state.size])
        action_index = self.get_action(state_as_dict,state)

        # Creates the action "object". If we use "print(action)" then it is readable.
        action = self.create_action_dict(self.gen_action_list, self.load_action_list, self.line_or_action_list, self.line_ex_action_list,
                                         action_index,None, None)
        # print(action)
        # self.action_list.append(action)
        return action

    def create_action_dict(self,gen_action_list, load_action_list, line_or_action_list, line_ex_action_list, action_index, episode, flag):
        action = self.action_space(
        {"change_bus": {"generators_id": gen_action_list[action_index], "loads_id": load_action_list[action_index],
                        "lines_or_id": line_or_action_list[action_index],
                        "lines_ex_id": line_ex_action_list[action_index]}})
        return action

    def get_action(self, state_as_dict, state):
        with self.sess.as_default():
            with self.sess.graph.as_default():
                # Predict the action using the internal neural network
                policy = self.actor.predict(state,batch_size=1).flatten()

        # Select first 4 best possible actions from the neural nets.
        policy_chosen_list = np.random.choice(self.action_size, 4, p=policy)

        # Simulate the impact of these actions and pick the best action that maximizes the reward.
        # (one-step lookahead).
        action_index = policy_chosen_list[0]
        action = self.create_action_dict(self.gen_action_list, self.load_action_list, self.line_or_action_list, self.line_ex_action_list, action_index, None, None)
        obs_0, rw_0, done_0, _  = state_as_dict.simulate(action)

        action_index = policy_chosen_list[1]
        action = self.create_action_dict(self.gen_action_list, self.load_action_list, self.line_or_action_list, self.line_ex_action_list, action_index, None, None)
        obs_1, rw_1, done_1, _  = state_as_dict.simulate(action)

        action_index = policy_chosen_list[2]
        action = self.create_action_dict(self.gen_action_list, self.load_action_list, self.line_or_action_list, self.line_ex_action_list, action_index, None, None)
        obs_2, rw_2, done_2, _  = state_as_dict.simulate(action)

        action_index = policy_chosen_list[3]
        action = self.create_action_dict(self.gen_action_list, self.load_action_list, self.line_or_action_list, self.line_ex_action_list, action_index, None, None)
        obs_3, rw_3, done_3, _  = state_as_dict.simulate(action)

        return policy_chosen_list[np.argmax([rw_0,rw_1,rw_2,rw_3])]

    # make loss function for Policy Gradient
    # [log(action probability) * advantages] will be input for the back prop
    # we add entropy of action probability to loss
    def actor_optimizer(self):
        action = K.placeholder(shape=(None, self.action_size))
        advantages = K.placeholder(shape=(None, ))

        policy = self.actor.output

        good_prob = K.sum(action * policy, axis=1)
        eligibility = K.log(good_prob + 1e-10) * K.stop_gradient(advantages)
        loss = -K.sum(eligibility)

        entropy = K.sum(policy * K.log(policy + 1e-10), axis=1)

        actor_loss = loss + 0.01*entropy

        optimizer = Adam(lr=self.actor_lr)
        # updates = optimizer.get_updates(params=self.actor.trainable_weights, constraints=[],loss=actor_loss)
        updates = optimizer.get_updates(params=self.actor.trainable_weights, loss=actor_loss)
        train = K.function([self.actor.input, action, advantages], tf.compat.v1.convert_to_tensor([]),updates=updates)
        return train

    # make loss function for Value approximation
    def critic_optimizer(self):
        discounted_reward = K.placeholder(shape=(None, ))

        value = self.critic.output

        loss = K.mean(K.square(discounted_reward - value))

        optimizer = Adam(lr=self.critic_lr)
        # updates = optimizer.get_updates(params=self.critic.trainable_weights, constraints=[],loss=loss)
        updates = optimizer.get_updates(params=self.critic.trainable_weights, loss=loss)
        train = K.function([self.critic.input, discounted_reward], tf.compat.v1.convert_to_tensor([]), updates=updates)
        return train

    # make agents(local) and start training
    def train(self, nn_weights_name):
        agents = [Agent(i, self.actor, self.critic, self.optimizer, self.env_name, self.discount_factor,
                        self.action_size, self.state_size, self.action_space_lists, self.profiles_chronics, self.sess) for i in range(self.threads)]

        for agent in agents:
            agent.start()

        while (len(scores) < EPISODES_train):
            time.sleep(200) # main thread saves the model every 200 sec
            print("len(scores) = ", len(scores))
            if (len(scores)>10):
                self.save_model(nn_weights_name,self.save_path)
                print("_______________________________________________________________________________________________________")
                print("saved NN model at episode", episode, "\n")
                print("_______________________________________________________________________________________________________")

    def load_model(self, nn_weight_name, load_path):
            self.actor.load_weights(os.path.join(load_path,nn_weight_name + "_actor.h5"))
            self.critic.load_weights(os.path.join(load_path,nn_weight_name + "_critic.h5"))

    def save_model(self, nn_weight_name, save_path):
        self.actor.save_weights(os.path.join(save_path,nn_weight_name + "_actor.h5"))
        self.critic.save_weights(os.path.join(save_path,nn_weight_name + "_critic.h5"))

# This is Agent(local) class for threading
class Agent(threading.Thread):
    def __init__(self, index, actor, critic, optimizer, env_name, discount_factor, action_size, state_size, action_space_lists,profiles_chronics,session):
        threading.Thread.__init__(self)

        self.states = []
        self.rewards = []
        self.actions = []

        self.index = index
        self.actor = actor
        self.critic = critic
        self.optimizer = optimizer
        self.env_name = env_name
        self.discount_factor = discount_factor
        self.action_size = action_size
        self.state_size = state_size
        self.session = session

        self.gen_action_list = action_space_lists[0]
        self.load_action_list = action_space_lists[1]
        self.line_or_action_list = action_space_lists[2]
        self.line_ex_action_list = action_space_lists[3]
        self.profiles_chronics = profiles_chronics

    # Thread interactive with environment
    def run(self):
        global episode
        episode = 0
        env = user_environment_make.set_environement(self.index,self.env_name,self.profiles_chronics)
        self.action_space = env.helper_action_player
        while episode < EPISODES_train:
            state = env.reset()
            state_as_dict = copy.deepcopy(state)
            # state = copy.deepcopy(np.reshape(state.to_vect(), [1, self.state_size]))
            state = useful_state(state_as_dict,env.backend.prod_pu_to_kv)
            state = state.reshape([1,state.size])
            score = 0
            time_step = 0
            max_action = 0
            non_zero_actions = 0
            epsilon = 0.5
            while True:
                # Decaying epsilon greedy. Not the best one. This agent needs a better exploration strategy to help
                # it learn to perform well.
                if np.random.random() < epsilon*(1/(episode/400+1)):
                    action_index = int(np.random.choice(self.action_size))
                    epison_flag = True
                else:
                    epison_flag = False
                    if time_step%1 == 0:# or max(state_as_dict.rho)>0.75:
                        action_index = self.get_action(state_as_dict,state)
                    else:
                        action_index = 0

                action = self.create_action_dict(self.gen_action_list, self.load_action_list, self.line_or_action_list, self.line_ex_action_list, action_index, episode, flag=1)

                # print(action)

                next_state, reward, done, flag = env.step(action)
                state_as_dict = copy.deepcopy(next_state)
                time_hour = state_as_dict.day*10000 + state_as_dict.hour_of_day * 100+ state_as_dict.minute_of_hour
                # next_state = np.reshape(next_state.to_vect(), [1, self.state_size]) if not done else np.zeros([1, self.state_size])
                next_state = useful_state(next_state,env.backend.prod_pu_to_kv)
                next_state = next_state.reshape([1,next_state.size])
                # next_state = observation_space.array_to_observation(next_state).as_minimalist().as_array()
                # score += (reward-0.1*(next_state[1]*next_state[1]+next_state[3]*next_state[3])) # reducing the reward based on speed...
                score += reward if not done else -100*(1+np.sqrt(episode)/10)
                non_zero_actions += 0 if action_index==0 else 1
                # if flag == None:
                #     self.memory(state, action, reward)
                # else:
                #     score -= 10 if flag.is_empty else 0
                #     self.memory(state, action, 0)
                self.memory(state, action_index, reward if not done else -100*(1+np.sqrt(episode)/10))

                state = copy.deepcopy(next_state) if not done else np.zeros([1, self.state_size])

                time_step += 1
                max_action = max(max_action,action_index)

                if done or time_step > time_step_end:
                    if done:
                        print("----STOPPED Thread:", self.index, "/ train episode: ", episode,  "/ instant reward",int(reward), "/ score : ", int(score),
                              "/ with final time:", time_step, "/ with final action", action_index,
                              "/Random action: ",epison_flag,"/ number of non-zero actions", non_zero_actions, "/ day_hour_min:", time_hour)
                    if time_step > time_step_end:
                        print("End Thread:", self.index, "/ train episode: ", episode,  "/ instant reward",int(reward), "/ score : ", int(score),
                              "/ with final time:", time_step, "/ with final action", action_index,
                              "/Random action: ",epison_flag,"/ number of non-zero actions", non_zero_actions, "/ day_hour_min:", time_hour)
                    # global scores
                    scores.append(score)
                    # print(len(scores))
                    # global episode
                    episode += 1
                    # if len(self.states) == 0:
                    #     k = 1
                    self.train_episode(True)  # max score = 80000
                    break
                else:
                    if time_step % 10 ==0:
                        print("Continue Thread:", self.index, "/ train episode: ", episode,  "/ instant reward",int(reward), "/ score : ", int(score),
                              "/ with recent time:", time_step, "/ with recent action", action_index,"/Random action: ",epison_flag,"/ number of non-zero actions", non_zero_actions, "/ max_action so far:", max_action)
                        self.train_episode(False)

    # In Policy Gradient, Q function is not available.
    # Instead agent uses sample returns for evaluating policy
    def discount_rewards(self, rewards, done=True):
        discounted_rewards = np.zeros_like(rewards)
        running_add = 0
        if not done:
            with self.session.as_default():
                with self.session.graph.as_default():
                    running_add = self.critic.predict(np.reshape(self.states[-1], (1, self.state_size)))[0]
        for t in reversed(range(0, len(rewards))):
            running_add = running_add * self.discount_factor + rewards[t]
            discounted_rewards[t] = running_add
        return discounted_rewards

    # save <s, a ,r> of each step
    # this is used for calculating discounted rewards
    def memory(self, state, action, reward):
        self.states.append(state[0])
        act = np.zeros(self.action_size)
        act[action] = 1
        self.actions.append(act)
        self.rewards.append(reward)

    # update policy network and value network every episode
    def train_episode(self, done):
        discounted_rewards = self.discount_rewards(self.rewards, done)
        with self.session.as_default():
            with self.session.graph.as_default():
                values = self.critic.predict(np.array(self.states))[0]
        values = np.reshape(values, len(values))

        advantages = discounted_rewards - values

        with self.session.as_default():
            with self.session.graph.as_default():
                state_as_tensor = tf.compat.v1.convert_to_tensor(np.asarray(self.states))
                action_as_tensor = tf.compat.v1.convert_to_tensor(np.asarray(self.actions))
                advantage_as_tensor = tf.compat.v1.convert_to_tensor(np.asarray(advantages))
                # reshaped_state_as_tensor = tf.compat.v1.reshape(state_as_tensor,self.actor.input.shape.dims)
                self.optimizer[0]([state_as_tensor, action_as_tensor, advantage_as_tensor])
                self.optimizer[1]([state_as_tensor, discounted_rewards])
                self.states, self.actions, self.rewards = [], [], []


    def get_action(self, state_as_dict, state):
        with self.session.as_default():
            with self.session.graph.as_default():
                # Predict the action using the internal neural network
                policy = self.actor.predict(state,batch_size=1).flatten()

        # Select first 4 best possible actions from the neural nets.
        policy_chosen_list = np.random.choice(self.action_size, 4, p=policy)

        # Simulate the impact of these actions and pick the best action that maximizes the reward.
        # (one-step lookahead).
        action_index = policy_chosen_list[0]
        action = self.create_action_dict(self.gen_action_list, self.load_action_list, self.line_or_action_list, self.line_ex_action_list, action_index, episode, flag=0)
        obs_0, rw_0, done_0, _  = state_as_dict.simulate(action)

        action_index = policy_chosen_list[1]
        action = self.create_action_dict(self.gen_action_list, self.load_action_list, self.line_or_action_list, self.line_ex_action_list, action_index, episode, flag=0)
        obs_1, rw_1, done_1, _  = state_as_dict.simulate(action)

        action_index = policy_chosen_list[2]
        action = self.create_action_dict(self.gen_action_list, self.load_action_list, self.line_or_action_list, self.line_ex_action_list, action_index, episode, flag=0)
        obs_2, rw_2, done_2, _  = state_as_dict.simulate(action)

        action_index = policy_chosen_list[3]
        action = self.create_action_dict(self.gen_action_list, self.load_action_list, self.line_or_action_list, self.line_ex_action_list, action_index, episode, flag=0)
        obs_3, rw_3, done_3, _  = state_as_dict.simulate(action)

        return policy_chosen_list[np.argmax([rw_0,rw_1,rw_2,rw_3])]

    def create_action_dict(self,gen_action_list, load_action_list, line_or_action_list, line_ex_action_list, action_index, episode, flag):
        action = self.action_space(
        {"change_bus": {"generators_id": gen_action_list[action_index], "loads_id": load_action_list[action_index],
                        "lines_or_id": line_or_action_list[action_index],
                        "lines_ex_id": line_ex_action_list[action_index]}})
        # if flag == 1:
        #     print("_______________________________________________________________________________________________________")
        #     print("thread number ", self.index," Executed action index in environment step:", action_index," at episode:", episode)
        #     # print(action)
        #     print("_______________________________________________________________________________________________________")
        # else:
        #     print("_______________________________________________________________________________________________________")
        #     print("thread number ", self.index," Executed action index at simulate step:", action_index," at episode:", episode)
        #     # print(action)
        #     print("_______________________________________________________________________________________________________")
        return action

# def set_environement(start_id,env_name,profiles_chronics):
#     param = Parameters()
#     param.NO_OVERFLOW_DISCONNECTION = True
#
#     env = make(env_name,chronics_path= profiles_chronics, reward_class=CombinedReward,param=param)
#     # Register custom reward for training
#     cr = env.reward_helper.template_reward
#     cr.addReward("overflow", CloseToOverflowReward(), 50.0)
#     cr.addReward("game", GameplayReward(), 100.0)
#     cr.initialize(env)
#
#     # Debug prints
#     print("Debug prints --->:")
#     print("Chronics location that being used:", env.chronics_handler.path)
#     print("Grid location being used:", env.init_grid_path)
#     print("Reward class that is being used:", env.rewardClass)
#     print("Action type class being used:", env.actionClass)
#     print("Observation type class being used:", env.observationClass)
#     print("Backend CSV file key names:", env.names_chronics_to_backend)
#     print("Legal action class being used:", env.legalActClass)
#     print("Voltage controller class being used:", env.voltagecontrolerClass)
#
#     if start_id != None:
#         env.chronics_handler.tell_id(start_id)
#         print("Thread number:",start_id,", ID of chronic current folder:",env.chronics_handler.real_data.id_chron_folder_current)
#     return env

# This below function reduces the size of the state space.
def useful_state(obs,value_multiplier):
    selected_obs = np.hstack((obs.topo_vect,obs.line_status))
    selected_obs = np.hstack((selected_obs,obs.load_p/100))#
    selected_obs = np.hstack((selected_obs,obs.load_q/100))
    selected_obs = np.hstack((selected_obs,obs.prod_p/100))
    selected_obs = np.hstack((selected_obs,obs.prod_v/value_multiplier))
    selected_obs = np.hstack((selected_obs,obs.rho))
    # selected_obs = np.hstack((selected_obs,obs.day))
    selected_obs = np.hstack((selected_obs,obs.hour_of_day/24))
    selected_obs = np.hstack((selected_obs,obs.minute_of_hour/60))
    # selected_obs = np.hstack((selected_obs,obs.day_of_week/7))
    return selected_obs







# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import numpy as np
import random
import tensorflow as tf
import tensorflow.keras as tfk
import tensorflow.keras.backend as K
import tensorflow.keras.models as tfkm
import tensorflow.keras.optimizers as tfko
import tensorflow.keras.layers as tfkl
import tensorflow.keras.activations as tfka


class SliceRDQN_NN(object):
    def __init__(self,
                 action_size,
                 observation_shape,
                 slices,
                 learning_rate = 1e-5):
        self.action_size = action_size
        self.observation_shape = observation_shape
        self.slices = slices
        self.n_slices = len(slices.keys())
        self.encoded_size = 384
        self.h_size = 512
        self.lr = learning_rate
        self.model = None
        self.construct_q_network()

    def forward_slice_encode(self, input_layer, slice_indexes, name):
        batch_size = tf.shape(input_layer)[0]
        trace_size = tf.shape(input_layer)[1]
        sliced = tf.gather(input_layer, slice_indexes, axis=2, name=name+"_slice")
        # Reshape to (batch_size * trace_len, data_size)
        # So that Dense process each data vect separately
        sliced = tf.reshape(sliced, (-1, sliced.shape[2] * sliced.shape[3]), name=name+"_enc_reshape")

        # Bayesian NN simulate using dropout
        lay1 = tfkl.Dropout(self.dropout_rate, name=name+"_bnn")(sliced)

        # Three layers encoder
        lay1 = tfkl.Dense(496, name=name+"_fc1")(lay1)
        lay1 = tf.nn.leaky_relu(lay1, alpha=0.01, name=name+"_leak_fc1")
        lay2 = tfkl.Dense(412, name=name+"_fc2")(lay1)
        lay2 = tf.nn.leaky_relu(lay2, alpha=0.01, name=name+"_leak_fc2")
        lay3 = tfkl.Dense(self.encoded_size, name=name+"_fc3")(lay2)
        lay3 = tf.nn.leaky_relu(lay3, alpha=0.01, name=name+"_leak_fc3")

        # Reshape encoded results to (batch_size, trace_len, encoded_size)
        encoded_shape = (batch_size, trace_size, self.encoded_size)
        encoded = tf.reshape(lay3, encoded_shape, name=name+"_encoded")
        return encoded

    def forward_slice_recur(self, input_rnn, mem_state, carry_state, name):
        # Single LSTM cell over trace_length
        lstm_layer = tfkl.LSTM(self.h_size, return_state=True,
                               name=name+"_lstm")
        states = [mem_state, carry_state]
        lstm_out, mem, carry = lstm_layer(input_rnn, initial_state=states)
        return lstm_out, mem, carry

    def forward_streams(self, hidden, q_len, name):
        # Advantage stream
        advantage = tfkl.Dense(64, name=name+"_fcadv")(hidden)
        advantage = tf.nn.leaky_relu(advantage, alpha=0.01, name=name+"_leakyadv")
        advantage = tfkl.Dense(q_len, name=name+"_adv")(advantage)
        advantage_mean = tf.math.reduce_mean(advantage, axis=1,
                                             keepdims=True,
                                             name=name+"_adv_mean")
        advantage = tfkl.subtract([advantage, advantage_mean],
                                  name= name+"_adv_sub")

        # Value stream
        value = tfkl.Dense(64, name=name+"_fcval")(hidden)
        value = tf.nn.leaky_relu(value, alpha=0.01, name=name+"_leakyval")
        value = tfkl.Dense(1, name=name+"_val")(value)

        # Q values = val + adv
        slice_q = tf.math.add(value, advantage, name=name+"_sliceq")
        return slice_q

    def construct_q_network(self):
        # Defines input tensors and scalars
        self.trace_length = tf.Variable(1, trainable=False,
                                        dtype=tf.int32, name="trace_len")
        self.dropout_rate = tf.Variable(0.0, trainable=False,
                                        dtype=tf.float32, name="drop_rate")
        states_shape = (self.n_slices, self.h_size)
        input_mem_states = tfk.Input(dtype=tf.float32, shape=states_shape,
                                     name='input_mem_states')
        input_carry_states = tfk.Input(dtype=tf.float32, shape=states_shape,
                                       name='input_carry_states')
        input_shape = (None,) + self.observation_shape
        input_layer = tfk.Input(dtype=tf.float32, shape=input_shape,
                                name='input_obs')

        # Forward encode slices
        slice_to_encoded = {}
        for slice_name, slice_v in self.slices.items():
            slice_idxs = slice_v["indexes"]
            encoded_slice = self.forward_slice_encode(input_layer,
                                                      slice_idxs,
                                                      slice_name)
            slice_to_encoded[slice_name] = encoded_slice

        # Forward recurring part
        slice_to_hidden = {}
        mem_states = []
        carry_states = []
        for idx, (slice_name, rnn_input) in enumerate(slice_to_encoded.items()):
            mem = input_mem_states[:, idx]
            carry = input_carry_states[:, idx]
            h, m, c = self.forward_slice_recur(rnn_input, mem, carry,
                                               slice_name)
            slice_to_hidden[slice_name] = h
            mem_states.append(m)
            carry_states.append(c)

        # Pack states outputs
        output_mem_states = tf.stack(mem_states, axis=1, name="mem_stack")
        output_carry_states = tf.stack(carry_states, axis=1, name="carry_stack")

        # Advantage and Value streams
        q_slices_li = []
        q_dn_li = []
        for slice_name, hidden in slice_to_hidden.items():
            q_len = self.slices[slice_name]["q_len"]
            slice_q = self.forward_streams(hidden, q_len, slice_name)
            q_slices_li.append(slice_q[:, 1:])
            q_dn_li.append(slice_q[:, :1])

        
        # Concatenate final Q
        q_dn = tf.stack(q_dn_li, 1, name="q_dn_stack")
        q_dn_mean = tf.reduce_mean(q_dn, axis=1, name="q_dn_mean")
        q_slices = tf.concat(q_slices_li, 1, name="q_slices_concat")
        output_q = tf.concat([q_dn_mean, q_slices], 1, name="q_batch")

        # Backwards pass
        model_inputs = [
            input_mem_states,
            input_carry_states,
            input_layer
        ]
        model_outputs = [
            output_q,
            output_mem_states,
            output_carry_states
        ]
        self.model = tfk.Model(inputs=model_inputs,
                               outputs=model_outputs,
                               name=self.__class__.__name__)
        losses = [
            self._clipped_mse_loss,
            self._no_loss,
            self._no_loss
        ]
        self.optimizer = tfko.Adam(lr=self.lr, clipnorm=1.0)
        self.model.compile(loss=losses, optimizer=self.optimizer)

    def _no_loss(self, y_true, y_pred):
        return 0.0

    def _clipped_mse_loss(self, Qnext, Q):
        loss = tf.math.reduce_mean(tf.math.square(Qnext - Q), name="loss_mse")
        clipped_loss = tf.clip_by_value(loss, 0.0, 1e3, name="loss_clip")
        return clipped_loss

    def bayesian_move(self, data, mem, carry, rate = 0.0):
        self.dropout_rate.assign(float(rate))
        self.trace_length.assign(1)

        input_shape = (1, 1) + self.observation_shape
        data_input = data.reshape(input_shape)
        mem_input = mem.reshape(1, self.n_slices, self.h_size)
        carry_input = carry.reshape(1, self.n_slices, self.h_size)
        model_input = [mem_input, carry_input, data_input]
        
        Q, mem, carry = self.model.predict(model_input, batch_size = 1)
        move = np.argmax(Q)

        return move, Q, mem, carry

    def random_move(self, data, mem, carry):
        self.trace_length.assign(1)
        self.dropout_rate.assign(0.0)

        input_shape = (1, 1) + self.observation_shape
        data_input = data.reshape(input_shape)
        mem_input = mem.reshape(1, self.n_slices, self.h_size)
        carry_input = carry.reshape(1, self.n_slices, self.h_size)
        model_input = [mem_input, carry_input, data_input]

        Q, mem, carry = self.model.predict(model_input, batch_size = 1) 
        move = np.random.randint(0, self.action_size)

        return move, mem, carry

    def predict_move(self, data, mem, carry):
        self.trace_length.assign(1)
        self.dropout_rate.assign(0.0)

        input_shape = (1, 1) + self.observation_shape
        data_input = data.reshape(input_shape)
        mem_input = mem.reshape(1, self.n_slices, self.h_size)
        carry_input = carry.reshape(1, self.n_slices, self.h_size)
        model_input = [mem_input, carry_input, data_input]
        
        Q, mem, carry = self.model.predict(model_input, batch_size = 1)
        move = np.argmax(Q)

        return move, Q, mem, carry

    def update_target_hard(self, target_model):
        this_weights = self.model.get_weights()
        target_model.set_weights(this_weights)

    def update_target_soft(self, target_model, tau=1e-2):
        tau_inv = 1.0 - tau
        # Get parameters to update
        target_params = target_model.trainable_variables
        main_params = self.model.trainable_variables

        # Update each param
        for i, var in enumerate(target_params):
            var_persist = var.value() * tau_inv
            var_update = main_params[i].value() * tau
            # Poliak averaging
            var.assign(var_update + var_persist)

    def save_network(self, path):
        # Saves model at specified path as h5 file
        # nothing has changed
        self.model.save_weights(path)
        print("Successfully saved model at: {}".format(path))

    def load_network(self, path):
        # nothing has changed
        self.model.load_weights(path)
        print("Successfully loaded network from: {}".format(path))




import os
import json

class DoubleDuelingDQNConfig():
    """
    DoubleDuelingDQN configurable hyperparameters
    exposed as class attributes
    """

    LR_DECAY_STEPS = 1024*64
    LR_DECAY_RATE = 0.95
    INITIAL_EPSILON = 0.99
    FINAL_EPSILON = 0.001
    DECAY_EPSILON = 1024*64
    DISCOUNT_FACTOR = 0.98
    PER_CAPACITY = 1024*64
    PER_ALPHA = 0.7
    PER_BETA = 0.5
    UPDATE_FREQ = 28
    UPDATE_TARGET_HARD_FREQ = -1
    UPDATE_TARGET_SOFT_TAU = 1e-3
    N_FRAMES = 4
    BATCH_SIZE = 32
    LR = 1e-5
    VERBOSE = True

    @staticmethod
    def from_json(json_in_path):
        with open(json_in_path, 'r') as fp:
            conf_json = json.load(fp)
        
        for k,v in conf_json.items():
            if hasattr(DoubleDuelingDQNConfig, k):
                setattr(DoubleDuelingDQNConfig, k, v)

    @staticmethod
    def to_json(json_out_path):
        conf_json = {}
        for attr in dir(DoubleDuelingDQNConfig):
            if attr.startswith('__') or callable(attr):
                continue
            conf_json[attr] = getattr(DoubleDuelingDQNConfig, attr)

        with open(json_out_path, 'w+') as fp:
            json.dump(fp, conf_json, indent=2)



#!/usr/bin/env python3

# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.


from l2rpn_baselines.Template.Template import Template


def train(env,
          name="Template",
          iterations=1,
          save_path=None,
          load_path=None,
          **kwargs):
    """
    This an example function to train a baseline.

    In order to be valid, if you chose (which is recommended) to provide a training script to help other retrain your
    baseline in different environments, or for longer period of time etc. This script should be contain the "train"
    function with at least the following arguments.

    Parameters
    ----------
    env: :class:`grid2op.Environment.Environment`
        The environmnent on which the baseline will be trained

    name: ``str``
        Fancy name you give to this baseline.

    iterations: ``int``
        Number of training iterations to perform

    save_path: ``str``
        The path where the baseline will be saved at the end of the training procedure.

    load_path: ``str``
        Path where to look for reloading the model. Use ``None`` if no model should be loaded.

    kwargs:
        Other key-word arguments that you might use for training.

    """

    baseline = Template(env.action_space,
                        env.observation_space,
                        name=name)

    if load_path is not None:
        baseline.load(load_path)

    baseline.train(env, iterations, save_path)
    # as in our example (and in our explanation) we recommend to save the mode regurlarly in the "train" function
    # it is not necessary to save it again here. But if you chose not to follow these advice, it is more than
    # recommended to save the "baseline" at the end of this function with:
    # baseline.save(path_save)


if __name__ == "__main__":
    """
    This is a possible implementation of the train script.
    """
    import grid2op
    from l2rpn_baselines.utils import cli_train
    args_cli = cli_train().parse_args()
    env = grid2op.make()
    train(env=env,
          name=args_cli.name,
          iterations=args_cli.num_train_steps,
          save_path=args_cli.save_path,
          load_path=args_cli.load_path)



try:
    import numpy as np
    import copy
    import os
    from grid2op.Parameters import Parameters
    from grid2op.Reward import L2RPNReward, CombinedReward, CloseToOverflowReward, GameplayReward
    import sys
    from l2rpn_baselines.AsynchronousActorCritic.AsynchronousActorCritic import A3CAgent
    from l2rpn_baselines.AsynchronousActorCritic.Action_reduced_list import main_function
    from l2rpn_baselines.AsynchronousActorCritic.user_environment_make import set_environement
    # import pytorch
except ImportError as exc_:
    raise ImportError("AsynchronousActorCritic baseline impossible to load the required dependencies for training the model. The error was: \n {}".format(exc_))

# from ActorCritic_Agent import *
# import Action_reduced_list
# import user_environment_make

# This below function reduces the size of the state space.
def useful_state(obs,value_multiplier):
    selected_obs = np.hstack((obs.topo_vect,obs.line_status))
    selected_obs = np.hstack((selected_obs,obs.load_p/100))#
    selected_obs = np.hstack((selected_obs,obs.load_q/100))
    selected_obs = np.hstack((selected_obs,obs.prod_p/100))
    selected_obs = np.hstack((selected_obs,obs.prod_v/value_multiplier))
    selected_obs = np.hstack((selected_obs,obs.rho))
    # selected_obs = np.hstack((selected_obs,obs.day))
    selected_obs = np.hstack((selected_obs,obs.hour_of_day/24))
    selected_obs = np.hstack((selected_obs,obs.minute_of_hour/60))
    # selected_obs = np.hstack((selected_obs,obs.day_of_week/7))
    return selected_obs

def train(env,name,iterations,save_path,load_path,env_name,profiles_chronics,time_step_end,Hyperparameters,Thread_count):
    """
    Parameters
        ----------
        env: :class:`grid2op.Environment.Environment`
            The environmnent on which the baseline will be trained
        name: ``str``
            Fancy name you give to this baseline that is being trained.
        iterations: ``int``
            Number of training iterations to perform
        save_path: ``str``
            The path where the baseline will be saved at the end of the training procedure.
        load_path: ``str``
            Path where to look for reloading the model. Use ``None`` if no model should be loaded.
        kwargs:
            Other key-word arguments that you might use for training.
    """

    # NB: "env" is not used. Instead the A3C agent will call function from "user_environment_make.py" in
    # "each unique thread" to create duplicate multiple environments in parallel. By no means this is perfect but this
    # works and we recommend using MultiEnv class from Grid2Op for future implementations that uses multiple threads.

    env_temp = set_environement(None, env_name, profiles_chronics)

    # Define the size of state space and action space.
    do_nothing_act = env_temp.helper_action_player({})

    obs, reward, done, info = env_temp.step(do_nothing_act)
    #conversion parameter
    value_multiplier = env_temp.backend.prod_pu_to_kv
    state_trimmed = useful_state(obs,value_multiplier)
    state_trimmed = state_trimmed.reshape([1,state_trimmed.size])
    state_size = state_trimmed.size

    # NOTE: Delete the .npy files if you are trying to solve for different bus system than previous solved power system
    # This code is implemented for a 14 bus system.

    # Reduce the action space and assign an index value to each unique possible action from the complete action space.
    gen_action_list, load_action_list, line_or_action_list, line_ex_action_list = main_function("14")
    action_size = load_action_list.__len__()
    action_space = env_temp.action_space
    del env_temp

    # Get the action space details as lists.
    action_space_lists = [gen_action_list, load_action_list, line_or_action_list, line_ex_action_list]

    # Create an agent architecture.
    global_agent = A3CAgent(state_size, action_size,env_name,action_space,value_multiplier,action_space_lists,
                            profiles_chronics,iterations,time_step_end,Hyperparameters,Thread_count,train_flag=True,
                            save_path=save_path)

    # Print model summary
    global_agent.actor.summary()
    global_agent.critic.summary()

    if load_path is not None:
        # Load the agent with trained neural network weights.
        try:
            global_agent.load_model(nn_weight_name=name,load_path=load_path)
            print("Loaded saved NN model parameters \n")
        except:
            print("Issue with loading the NN weights/No existing model is found or saved model sizes do not match. Exiting...\n")
            exit()
    else:
        print("Training the A3C agent from scratch i.e., without loading the previously trained neural network weights.")

    global_agent.train(name)

if __name__ == "__main__":
    # Name of the ".h5" files that stores the actor and critic neural network weights.
    name = "grid2op_14_a3c"
    # Total number of episodes to train the A3C agent.
    EPISODES_train = 500
    # Maximum number of time steps or iterations taken in each episode during the training.
    time_step_end = 3000
    # Number of parallel workers or threads to train the A3C agent.
    Thread_count = 7
    # location where ".h5" files to be saved.
    save_path = os.path.join(os.getcwd(),"nn_weight_folder")
    # location to load pre-trained ".h5" files (neural network weights) OR provide "None" as input.
    load_path = None #os.path.join(os.getcwd(),"nn_weight_folder")

    # Neural network architecture hyper parameters. For more details on structure see the ".summary()" that will be
    # printed when running this code.
    Hyperparameters = {}
    Hyperparameters["actor_learning_rate"] = 0.0003
    Hyperparameters["critic_learning_rate"] = 0.0003
    Hyperparameters["discount_factor"] = 0.95
    Hyperparameters["size_of_hidden_layer_1"] = 200
    Hyperparameters["size_of_hidden_layer_2"] = 100

    # Create the environment
    # env = user_environment_make.set_environement(None, env_name, profiles_chronics)

    # Name of the Grid2Op environment to train the A3C RL agent on.
    env_name = 'l2rpn_case14_sandbox'
    # Location of the chronic files. Change this to your chronics location.
    profiles_chronics = r"C:\Users\kisha\data_grid2op\l2rpn_case14_sandbox\chronics"
    print("_____________________________________________")
    print("NOTE: PLEASE MAKE CHANGES TO THE FILE user_environment_make.py TO MAKE YOUR ENVIRONMENT. This "
          "user_environment_make.py will be called wherever it is necessary when training the A3C agent")
    print("_____________________________________________")

    # Train function that trains the A3C agent.
    train(env=None,name=name,iterations=EPISODES_train,save_path=save_path,load_path=load_path,
          env_name=env_name,profiles_chronics=profiles_chronics,time_step_end=time_step_end,
          Hyperparameters=Hyperparameters,Thread_count=Thread_count)





# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.
import os
import copy

from l2rpn_baselines.utils import NNParam
from l2rpn_baselines.DuelQSimple.DuelQ_NN import DuelQ_NN


class DuelQ_NNParam(NNParam):
    _int_attr = copy.deepcopy(NNParam._int_attr)
    _float_attr = copy.deepcopy(NNParam._float_attr)
    _str_attr = copy.deepcopy(NNParam._str_attr)
    _list_float = copy.deepcopy(NNParam._list_float)
    _list_str = copy.deepcopy(NNParam._list_str)
    _list_int = copy.deepcopy(NNParam._list_int)

    nn_class = DuelQ_NN

    def __init__(self,
                 action_size,
                 observation_size,  # TODO this might not be usefull
                 sizes,
                 activs,
                 list_attr_obs
                 ):
        NNParam.__init__(self,
                         action_size,
                         observation_size,  # TODO this might not be usefull
                         sizes,
                         activs,
                         list_attr_obs
                         )


# Modified from
# https://github.com/openai/baselines

import numpy as np
import random

from l2rpn_baselines.DoubleDuelingDQN.segment_tree import SumSegmentTree, MinSegmentTree


class BaseReplayBuffer(object):
    def __init__(self, size):
        """Create Replay buffer.

        Parameters
        ----------
        size: int
            Max number of transitions to store in the buffer. When the buffer
            overflows the old memories are dropped.
        """
        self._storage = []
        self._maxsize = size
        self._next_idx = 0

    def __len__(self):
        return len(self._storage)

    def add(self, obs_t, action, reward, obs_tp1, done):
        data = (obs_t, action, reward, obs_tp1, done)

        if self._next_idx >= len(self._storage):
            self._storage.append(data)
        else:
            self._storage[self._next_idx] = data
        self._next_idx = (self._next_idx + 1) % self._maxsize

    def _encode_sample(self, idxes):
        obses_t, actions, rewards, obses_tp1, dones = [], [], [], [], []
        for i in idxes:
            data = self._storage[i]
            obs_t, action, reward, obs_tp1, done = data
            obses_t.append(np.array(obs_t, copy=False))
            actions.append(np.array(action, copy=False))
            rewards.append(reward)
            obses_tp1.append(np.array(obs_tp1, copy=False))
            dones.append(done)
        return np.array(obses_t), np.array(actions), np.array(rewards), np.array(obses_tp1), np.array(dones)

    def sample(self, batch_size):
        """Sample a batch of experiences.

        Parameters
        ----------
        batch_size: int
            How many transitions to sample.

        Returns
        -------
        obs_batch: np.array
            batch of observations
        act_batch: np.array
            batch of actions executed given obs_batch
        rew_batch: np.array
            rewards received as results of executing act_batch
        next_obs_batch: np.array
            next set of observations seen after executing act_batch
        done_mask: np.array
            done_mask[i] = 1 if executing act_batch[i] resulted in
            the end of an episode and 0 otherwise.
        """
        idxes = [random.randint(0, len(self._storage) - 1) for _ in range(batch_size)]
        return self._encode_sample(idxes)


class PrioritizedReplayBuffer(BaseReplayBuffer):
    def __init__(self, size, alpha):
        """Create Prioritized Replay buffer.

        Parameters
        ----------
        size: int
            Max number of transitions to store in the buffer. When the buffer
            overflows the old memories are dropped.
        alpha: float
            how much prioritization is used
            (0 - no prioritization, 1 - full prioritization)

        See Also
        --------
        ReplayBuffer.__init__
        """
        super(PrioritizedReplayBuffer, self).__init__(size)
        assert alpha >= 0
        self._alpha = alpha

        it_capacity = 1
        while it_capacity < size:
            it_capacity *= 2

        self._it_sum = SumSegmentTree(it_capacity)
        self._it_min = MinSegmentTree(it_capacity)
        self._max_priority = 1.0

    def add(self, *args, **kwargs):
        """See ReplayBuffer.store_effect"""
        idx = self._next_idx
        super().add(*args, **kwargs)
        self._it_sum[idx] = self._max_priority ** self._alpha
        self._it_min[idx] = self._max_priority ** self._alpha

    def _sample_proportional(self, batch_size):
        res = []
        p_total = self._it_sum.sum(0, len(self._storage) - 1)
        every_range_len = p_total / batch_size
        for i in range(batch_size):
            mass = random.random() * every_range_len + i * every_range_len
            idx = self._it_sum.find_prefixsum_idx(mass)
            res.append(idx)
        return res

    def sample(self, batch_size, beta):
        """Sample a batch of experiences.

        compared to ReplayBuffer.sample
        it also returns importance weights and idxes
        of sampled experiences.


        Parameters
        ----------
        batch_size: int
            How many transitions to sample.
        beta: float
            To what degree to use importance weights
            (0 - no corrections, 1 - full correction)

        Returns
        -------
        obs_batch: np.array
            batch of observations
        act_batch: np.array
            batch of actions executed given obs_batch
        rew_batch: np.array
            rewards received as results of executing act_batch
        next_obs_batch: np.array
            next set of observations seen after executing act_batch
        done_mask: np.array
            done_mask[i] = 1 if executing act_batch[i] resulted in
            the end of an episode and 0 otherwise.
        weights: np.array
            Array of shape (batch_size,) and dtype np.float32
            denoting importance weight of each sampled transition
        idxes: np.array
            Array of shape (batch_size,) and dtype np.int32
            idexes in buffer of sampled experiences
        """
        assert beta > 0

        idxes = self._sample_proportional(batch_size)

        weights = []
        p_min = self._it_min.min() / self._it_sum.sum()
        max_weight = (p_min * len(self._storage)) ** (-beta)

        for idx in idxes:
            p_sample = self._it_sum[idx] / self._it_sum.sum()
            weight = (p_sample * len(self._storage)) ** (-beta)
            weights.append(weight / max_weight)
        weights = np.array(weights)
        encoded_sample = self._encode_sample(idxes)
        return tuple(list(encoded_sample) + [weights, idxes])

    def update_priorities(self, idxes, priorities):
        """Update priorities of sampled transitions.

        sets priority of transition at index idxes[i] in buffer
        to priorities[i].

        Parameters
        ----------
        idxes: [int]
            List of idxes of sampled transitions
        priorities: [float]
            List of updated priorities corresponding to
            transitions at the sampled idxes denoted by
            variable `idxes`.
        """
        assert len(idxes) == len(priorities)
        for idx, priority in zip(idxes, priorities):
            assert priority > 0
            assert 0 <= idx < len(self._storage)
            self._it_sum[idx] = priority ** self._alpha
            self._it_min[idx] = priority ** self._alpha

            self._max_priority = max(self._max_priority, priority)



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import os
from abc import ABC, abstractmethod
import numpy as np
import warnings
import tensorflow as tf
import tensorflow.keras.optimizers as tfko

from tensorflow.keras.models import load_model

from l2rpn_baselines.utils.TrainingParam import TrainingParam


# refactorization of the code in a base class to avoid copy paste.
class BaseDeepQ(ABC):
    """
    This class aims at representing the Q value (or more in case of SAC) parametrization by
    a neural network.

    It is composed of 2 different networks:

    - model: which is the main model
    - target_model: which has the same architecture and same initial weights as "model" but is updated less frequently
      to stabilize training

    It has basic methods to make predictions, to train the model, and train the target model.

    This class is abstraction and need to be overide in order to create object from this class. The only pure virtual
    function is :func:`BaseDeepQ.construct_q_network` that creates the neural network from the nn_params
    (:class:`NNParam`) provided as input

    Attributes
    ----------
    _action_size: ``int``
        Total number of actions

    _observation_size: ``int``
        Size of the observation space considered

    _nn_archi: :class:`NNParam`
        The parameters of the neural networks that will be created

    _training_param: :class:`TrainingParam`
        The meta parameters for the training scheme (used especially for learning rate or gradient clipping for example)

    _lr: ``float``
        The  initial learning rate

    _lr_decay_steps: ``float``
        The decay step of the learning rate

    _lr_decay_rate: ``float``
        The rate at which the learning rate will decay

    _model:
        Main neural network model, here a keras Model object.

    _target_model:
        a copy of the main neural network that will be updated less frequently (also known as "target model" in RL
        community)


    """

    def __init__(self,
                 nn_params,
                 training_param=None,
                 verbose=False):
        self._action_size = nn_params.action_size
        self._observation_size = nn_params.observation_size
        self._nn_archi = nn_params
        self.verbose = verbose

        if training_param is None:
            self._training_param = TrainingParam()
        else:
            self._training_param = training_param

        self._lr = training_param.lr
        self._lr_decay_steps = training_param.lr_decay_steps
        self._lr_decay_rate = training_param.lr_decay_rate

        self._model = None
        self._target_model = None
        self._schedule_model = None
        self._optimizer_model = None
        self._custom_objects = None  # to be able to load other keras layers type

    def make_optimiser(self):
        """
        helper function to create the proper optimizer (Adam) with the learning rates and its decay
        parameters.
        """
        schedule = tfko.schedules.InverseTimeDecay(self._lr, self._lr_decay_steps, self._lr_decay_rate)
        return schedule, tfko.Adam(learning_rate=schedule)

    @abstractmethod
    def construct_q_network(self):
        """
         Abstract method that need to be overide.

         It should create :attr:`BaseDeepQ._model` and :attr:`BaseDeepQ._target_model`
        """
        raise NotImplementedError("Not implemented")

    def predict_movement(self, data, epsilon, batch_size=None):
        """
        Predict movement of game controler where is epsilon probability randomly move."""
        if batch_size is None:
            batch_size = data.shape[0]

        rand_val = np.random.random(batch_size)
        q_actions = self._model.predict(data, batch_size=batch_size)

        opt_policy = np.argmax(np.abs(q_actions), axis=-1)
        opt_policy[rand_val < epsilon] = np.random.randint(0, self._action_size, size=(np.sum(rand_val < epsilon)))
        return opt_policy, q_actions[0, opt_policy]

    def train(self, s_batch, a_batch, r_batch, d_batch, s2_batch, tf_writer=None, batch_size=None):
        """Trains network to fit given parameters"""
        if batch_size is None:
            batch_size = s_batch.shape[0]

        # Save the graph just the first time
        if tf_writer is not None:
            tf.summary.trace_on()
        targets = self._model.predict(s_batch, batch_size=batch_size)
        if tf_writer is not None:
            with tf_writer.as_default():
                tf.summary.trace_export("model-graph", 0)
            tf.summary.trace_off()
        fut_action = self._target_model.predict(s2_batch, batch_size=batch_size)

        targets[:, a_batch.flatten()] = r_batch
        targets[d_batch, a_batch[d_batch]] += self._training_param.discount_factor * np.max(fut_action[d_batch], axis=-1)

        loss = self.train_on_batch(self._model, self._optimizer_model, s_batch, targets)
        return loss

    def train_on_batch(self, model, optimizer_model, x, y_true):
        """train the model on a batch of example. This can be overide"""
        loss = model.train_on_batch(x, y_true)
        return loss

    @staticmethod
    def get_path_model(path, name=None):
        """
        Get the location at which the neural networks will be saved.

        Returns
        -------
        path_model: ``str``
            The path at which the model will be saved (path include both path and name, it is the full path at which
            the neural networks are saved)

        path_target_model: ``str``
            The path at which the target model will be saved
        """
        if name is None:
            path_model = path
        else:
            path_model = os.path.join(path, name)
        path_target_model = "{}_target".format(path_model)
        return path_model, path_target_model

    def save_network(self, path, name=None, ext="h5"):
        """
        save the neural networks.

        Parameters
        ----------
        path: ``str``
            The path at which the models need to be saved
        name: ``str``
            The name given to this model

        ext: ``str``
            The file extension (by default h5)
        """
        # Saves model at specified path as h5 file
        # nothing has changed
        path_model, path_target_model = self.get_path_model(path, name)
        self._model.save('{}.{}'.format(path_model, ext))
        self._target_model.save('{}.{}'.format(path_target_model, ext))

    def load_network(self, path, name=None, ext="h5"):
        """
        Load the neural networks.
        Parameters
        ----------
        path: ``str``
            The path at which the models need to be saved
        name: ``str``
            The name given to this model

        ext: ``str``
            The file extension (by default h5)
        """
        path_model, path_target_model = self.get_path_model(path, name)
        self._model = load_model('{}.{}'.format(path_model, ext), custom_objects=self._custom_objects)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            self._target_model = load_model('{}.{}'.format(path_target_model, ext), custom_objects=self._custom_objects)
        if self.verbose:
            print("Succesfully loaded network.")

    def target_train(self):
        """
        update the target model with the parameters given in the :attr:`BaseDeepQ._training_param`.
        """
        # nothing has changed from the original implementation
        model_weights = self._model.get_weights()
        target_model_weights = self._target_model.get_weights()
        for i in range(len(model_weights)):
            target_model_weights[i] = self._training_param.tau * model_weights[i] + (1 - self._training_param.tau) * \
                                      target_model_weights[i]
        self._target_model.set_weights(target_model_weights)


# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import os
import numpy as np
from tqdm import tqdm
import tensorflow as tf

import grid2op
from grid2op.Exceptions import Grid2OpException
from grid2op.Agent import AgentWithConverter
from grid2op.Converter import IdToAct

from l2rpn_baselines.utils.ReplayBuffer import ReplayBuffer
from l2rpn_baselines.utils.TrainingParam import TrainingParam

try:
    from grid2op.Chronics import MultifolderWithCache
    _CACHE_AVAILABLE_DEEPQAGENT = True
except ImportError:
    _CACHE_AVAILABLE_DEEPQAGENT = False


class DeepQAgent(AgentWithConverter):
    """
    This class allows to train and log the training of different Q learning algorithm.

    It is not meant to be the state of the art implement of some baseline. It is rather meant to be a set of
    useful functions that allows to easily develop an environment if we want to get started in RL using grid2op.

    It derives from :class:`grid2op.Agent.AgentWithConverter` and as such implements the :func:`DeepQAgent.convert_obs`
    and :func:`DeepQAgent.my_act`

    It is suppose to be a Baseline, so it implements also the

    - :func:`DeepQAgent.load`: to load the agent
    - :func:`DeepQAgent.save`: to save the agent
    - :func:`DeepQAgent.train`: to train the agent

    TODO description of the training scheme!

    Attributes
    ----------
    filter_action_fun: ``callable``
        The function used to filter the action of the action space. See the documentation of grid2op:
        :class:`grid2op.Converter.IdToAct`
        `here <https://grid2op.readthedocs.io/en/v0.9.3/converter.html#grid2op.Converter.IdToAct>`_ for more
        information.

    replay_buffer:
        The experience replay buffer

    deep_q: :class:`BaseDeepQ`
        The neural network, represented as a :class:`BaseDeepQ` object.

    name: ``str``
        The name of the Agent

    store_action: ``bool``
        Whether you want to register which action your agent took or not. Saving the action can slow down a bit
        the computation (less than 1%) but can help understand what your agent is doing during its learning process.

    dict_action: ``str``
        The action taken by the agent, represented as a dictionnary. This can be useful to know which type of actions
        is taken by your agent. Only filled if :attr:DeepQAgent.store_action` is ``True``

    istraining: ``bool``
        Whether you are training this agent or not. No more really used. Mainly used for backward compatibility.

    epsilon: ``float``
        The epsilon greedy exploration parameter.

    nb_injection: ``int``
        Number of action tagged as "injection". See the
        `official grid2op documentation <https://grid2op.readthedocs.io/en/v0.9.3/action.html?highlight=get_types#grid2op.Action.BaseAction.get_types>`_
        for more information.

    nb_voltage: ``int``
        Number of action tagged as "voltage". See the
        `official grid2op documentation <https://grid2op.readthedocs.io/en/v0.9.3/action.html?highlight=get_types#grid2op.Action.BaseAction.get_types>`_
        for more information.

    nb_topology: ``int``
        Number of action tagged as "topology". See the
        `official grid2op documentation <https://grid2op.readthedocs.io/en/v0.9.3/action.html?highlight=get_types#grid2op.Action.BaseAction.get_types>`_
        for more information.

    nb_redispatching: ``int``
        Number of action tagged as "redispatching". See the
        `official grid2op documentation <https://grid2op.readthedocs.io/en/v0.9.3/action.html?highlight=get_types#grid2op.Action.BaseAction.get_types>`_
        for more information.

    nb_do_nothing: ``int``
        Number of action tagged as "do_nothing", *ie* when an action is not modifiying the state of the grid. See the
        `official grid2op documentation <https://grid2op.readthedocs.io/en/v0.9.3/action.html?highlight=get_types#grid2op.Action.BaseAction.get_types>`_
        for more information.

    verbose: ``bool``
        An effort will be made on the logging (outside of trensorboard) of the training. For now: verbose=True will
        allow some printing on the command prompt, and verbose=False will drastically reduce the amount of information
        printed during training.

    """
    def __init__(self,
                 action_space,
                 nn_archi,
                 name="DeepQAgent",
                 store_action=True,
                 istraining=False,
                 filter_action_fun=None,
                 verbose=False,
                 **kwargs_converters):
        AgentWithConverter.__init__(self, action_space, action_space_converter=IdToAct, **kwargs_converters)
        self.filter_action_fun = filter_action_fun
        if self.filter_action_fun is not None:
            self.action_space.filter_action(self.filter_action_fun)

        # and now back to the origin implementation
        self.replay_buffer = None
        self.__nb_env = None

        self.deep_q = None
        self._training_param = None
        self._tf_writer = None
        self.name = name
        self._losses = None
        self.__graph_saved = False
        self.store_action = store_action
        self.dict_action = {}
        self.istraining = istraining
        self._actions_per_1000steps = np.zeros((1000, self.action_space.size()), dtype=np.int)
        self._illegal_actions_per_1000steps = np.zeros(1000, dtype=np.int)
        self._ambiguous_actions_per_1000steps = np.zeros(1000, dtype=np.int)
        self.epsilon = 1.0

        # for tensorbaord
        self._train_lr = None

        self._reset_num = None

        self._max_iter_env_ = 1000000
        self._curr_iter_env = 0
        self._max_reward = 0.

        # action type
        self.nb_injection = 0
        self.nb_voltage = 0
        self.nb_topology = 0
        self.nb_line = 0
        self.nb_redispatching = 0
        self.nb_do_nothing = 0

        # for over sampling the hard scenarios
        self._prev_obs_num = 0
        self._time_step_lived = None
        self._nb_chosen = None
        self._proba = None
        self._prev_id = 0
        # this is for the "limit the episode length" depending on your previous success
        self._total_sucesses = 0

        # update frequency of action types
        self._nb_updated_act_tensorboard = None

        # neural network architecture
        self._nn_archi = nn_archi

        # observation tranformers
        self._obs_as_vect = None
        self._tmp_obs = None
        self._indx_obs = None
        self.verbose = verbose

    # grid2op.Agent interface
    def convert_obs(self, observation):
        """
        Generic way to convert an observation. This transform it to a vector and the select the attributes that were
        selected in :attr:`l2rpn_baselines.utils.NNParams.list_attr_obs` (that have been extracted once and for all
        in the :attr:`DeepQAgent._indx_obs` vector).

        Parameters
        ----------
        observation: :class:`grid2op.Observation.BaseObservation`
            The current observation sent by the environment

        Returns
        -------
        _tmp_obs: ``numpy.ndarray``
            The observation as vector with only the proper attribute selected (TODO scaling will be available
            in future version)

        """
        obs_as_vect = observation.to_vect()
        self._tmp_obs[:] = obs_as_vect[self._indx_obs]
        return self._tmp_obs

    def my_act(self, transformed_observation, reward, done=False):
        """
        This function will return the action (its id) selected by the underlying :attr:`DeepQAgent.deep_q` network.

        Before being used, this method require that the :attr:`DeepQAgent.deep_q` is created. To that end a call
        to :func:`DeepQAgent.init_deep_q` needs to have been performed (this is automatically done if you use
        baseline we provide and their `evaluate` and `train` scripts).

        Parameters
        ----------
        transformed_observation: ``numpy.ndarray``
            The observation, as transformed after :func:`DeepQAgent.convert_obs`

        reward: ``float``
            The reward of the last time step. Ignored by this method. Here for retro compatibility with openAI
            gym interface.

        done: ``bool``
            Whether the episode is over or not. This is not used, and is only present to be compliant with
            open AI gym interface

        Returns
        -------
        res: ``int``
            The id the action taken.

        """
        predict_movement_int, *_ = self.deep_q.predict_movement(transformed_observation, epsilon=0.0)
        res = int(predict_movement_int)
        self._store_action_played(res)
        return res

    @staticmethod
    def get_action_size(action_space, filter_fun, kwargs_converters):
        """
        This function allows to get the size of the action space if we were to built a :class:`DeepQAgent`
        with this parameters.

        Parameters
        ----------
        action_space: :class:`grid2op.ActionSpace`
            The grid2op action space used.

        filter_fun: ``callable``
            see :attr:`DeepQAgent.filter_fun` for more information

        kwargs_converters: ``dict``
            see the documentation of grid2op for more information:
            `here <https://grid2op.readthedocs.io/en/v0.9.3/converter.html?highlight=idToAct#grid2op.Converter.IdToAct.init_converter>`_


        See Also
        --------
            The official documentation of grid2Op, and especially its class "IdToAct" at this address
            `IdToAct <https://grid2op.readthedocs.io/en/v0.9.3/converter.html?highlight=idToAct#grid2op.Converter.IdToAct>`_

        """
        converter = IdToAct(action_space)
        converter.init_converter(**kwargs_converters)
        if filter_fun is not None:
            converter.filter_action(filter_fun)
        return converter.n

    def init_obs_extraction(self, env):
        """
        This method should be called to initialize the observation (feed as a vector in the neural network)
        from its description as a list of its attribute names.
        """
        tmp = np.zeros(0, dtype=np.uint)  # TODO platform independant
        for obs_attr_name in self._nn_archi.get_obs_attr():
            beg_, end_, dtype_ = env.observation_space.get_indx_extract(obs_attr_name)
            tmp = np.concatenate((tmp, np.arange(beg_, end_, dtype=np.uint)))
        self._indx_obs = tmp
        self._tmp_obs = np.zeros((1, tmp.shape[0]), dtype=np.float32)

    # baseline interface
    def load(self, path):
        """
        Part of the l2rpn_baselines interface, this function allows to read back a trained model, to continue the
        training or to evaluate its performance for example.

        **NB** To reload an agent, it must have exactly the same name and have been saved at the right location.

        Parameters
        ----------
        path: ``str``
            The path where the agent has previously beens saved.

        """
        # not modified compare to original implementation
        tmp_me = os.path.join(path, self.name)
        if not os.path.exists(tmp_me):
            raise RuntimeError("The model should be stored in \"{}\". But this appears to be empty".format(tmp_me))
        self._load_action_space(tmp_me)

        # TODO handle case where training param class has been overidden
        self._training_param = TrainingParam.from_json(os.path.join(tmp_me, "training_params.json".format(self.name)))
        self.deep_q = self._nn_archi.make_nn(self._training_param)
        try:
            self.deep_q.load_network(tmp_me, name=self.name)
        except Exception as e:
            raise RuntimeError("Impossible to load the model located at \"{}\" with error \n{}".format(path, e))

        for nm_attr in ["_time_step_lived", "_nb_chosen", "_proba"]:
            conv_path = os.path.join(tmp_me, "{}.npy".format(nm_attr))
            if os.path.exists(conv_path):
                setattr(self, nm_attr, np.load(file=conv_path))

    def save(self, path):
        """
        Part of the l2rpn_baselines interface, this allows to save a model. Its name is used at saving time. The
        same name must be reused when loading it back.

        Parameters
        ----------
        path: ``str``
            The path where to save the agent.

        """
        if path is not None:
            tmp_me = os.path.join(path, self.name)
            if not os.path.exists(tmp_me):
                os.mkdir(tmp_me)
            nm_conv = "action_space.npy"
            conv_path = os.path.join(tmp_me, nm_conv)
            if not os.path.exists(conv_path):
                self.action_space.save(path=tmp_me, name=nm_conv)

            self._training_param.save_as_json(tmp_me, name="training_params.json")
            self._nn_archi.save_as_json(tmp_me, "nn_architecture.json")
            self.deep_q.save_network(tmp_me, name=self.name)

            # TODO save the "oversampling" part, and all the other info
            for nm_attr in ["_time_step_lived", "_nb_chosen", "_proba"]:
                conv_path = os.path.join(tmp_me, "{}.npy".format(nm_attr))
                attr_ = getattr(self, nm_attr)
                if attr_ is not None:
                    np.save(arr=attr_, file=conv_path)

    def train(self,
              env,
              iterations,
              save_path,
              logdir,
              training_param=None):
        """
        Part of the public l2rpn-baselines interface, this function allows to train the baseline.

        If `save_path` is not None, the the model is saved regularly, and also at the end of training.

        TODO explain a bit more how you can train it.

        Parameters
        ----------
        env: :class:`grid2op.Environment.Environment` or :class:`grid2op.Environment.MultiEnvironment`
            The environment used to train your model.

        iterations: ``int``
            The number of training iteration. NB when reloading a model, this is **NOT** the training steps that will
            be used when re training. Indeed, if `iterations` is 1000 and the model was already trained for 750 time
            steps, then when reloaded, the training will occur on 250 (=1000 - 750) time steps only.

        save_path: ``str``
            Location at which to save the model

        logdir: ``str``
            Location at which tensorboard related information will be kept.

        training_param: :class:`l2rpn_baselines.utils.TrainingParam`
            The meta parameters for the training procedure. This is currently ignored if the model is reloaded (in that
            case the parameters used when first created will be used)

        """

        if training_param is None:
            training_param = TrainingParam()

        self._train_lr = training_param.lr

        if self._training_param is None:
            self._training_param = training_param
        else:
            training_param = self._training_param
        self._init_deep_q(self._training_param, env)

        self._init_replay_buffer()

        # efficient reading of the data (read them by chunk of roughly 1 day
        nb_ts_one_day = 24 * 60 / 5  # number of time steps per day
        self._set_chunk(env, nb_ts_one_day)

        # Create file system related vars
        if save_path is not None:
            save_path = os.path.abspath(save_path)
            os.makedirs(save_path, exist_ok=True)

        if logdir is not None:
            logpath = os.path.join(logdir, self.name)
            self._tf_writer = tf.summary.create_file_writer(logpath, name=self.name)
        else:
            logpath = None
            self._tf_writer = None
        UPDATE_FREQ = training_param.update_tensorboard_freq  # update tensorboard every "UPDATE_FREQ" steps
        SAVING_NUM = training_param.save_model_each

        if isinstance(env, grid2op.Environment.Environment):
            self.__nb_env = 1
        else:
            import warnings
            nb_env = env.nb_env
            warnings.warn("Training using {} environments".format(nb_env))
            self.__nb_env = nb_env

        self.init_obs_extraction(env)

        training_step = self._training_param.last_step

        # some parameters have been move to a class named "training_param" for convenience
        self.epsilon = self._training_param.initial_epsilon

        # now the number of alive frames and total reward depends on the "underlying environment". It is vector instead
        # of scalar
        alive_frame, total_reward = self._init_global_train_loop()
        reward, done = self._init_local_train_loop()
        epoch_num = 0
        self._losses = np.zeros(iterations)
        alive_frames = np.zeros(iterations)
        total_rewards = np.zeros(iterations)
        new_state = None
        self._reset_num = 0
        self._curr_iter_env = 0
        self._max_reward = env.reward_range[1]

        # action types
        # injection, voltage, topology, line, redispatching = action.get_types()
        self.nb_injection = 0
        self.nb_voltage = 0
        self.nb_topology = 0
        self.nb_line = 0
        self.nb_redispatching = 0
        self.nb_do_nothing = 0

        # for non uniform random sampling of the scenarios
        th_size = None
        self._prev_obs_num = 0
        if self.__nb_env == 1:
            # TODO make this available for multi env too
            if _CACHE_AVAILABLE_DEEPQAGENT:
                if isinstance(env.chronics_handler.real_data, MultifolderWithCache):
                    th_size = env.chronics_handler.real_data.cache_size
            if th_size is None:
                th_size = len(env.chronics_handler.real_data.subpaths)

            # number of time step lived per possible scenarios
            if self._time_step_lived is None or self._time_step_lived.shape[0] != th_size:
                self._time_step_lived = np.zeros(th_size, dtype=np.uint64)
            # number of time a given scenario has been played
            if self._nb_chosen is None or self._nb_chosen.shape[0] != th_size:
                self._nb_chosen = np.zeros(th_size, dtype=np.uint)
            # number of time a given scenario has been played
            if self._proba is None or self._proba.shape[0] != th_size:
                self._proba = np.ones(th_size, dtype=np.float64)

        self._prev_id = 0
        # this is for the "limit the episode length" depending on your previous success
        self._total_sucesses = 0
        # update the frequency of action types
        self._nb_updated_act_tensorboard = 0

        with tqdm(total=iterations - training_step, disable=not self.verbose) as pbar:
            while training_step < iterations:
                # reset or build the environment
                initial_state = self._need_reset(env, training_step, epoch_num, done, new_state)

                # Slowly decay the exploration parameter epsilon
                # if self.epsilon > training_param.FINAL_EPSILON:
                self.epsilon = self._training_param.get_next_epsilon(current_step=training_step)

                # then we need to predict the next moves. Agents have been adapted to predict a batch of data
                pm_i, pq_v, act = self._next_move(initial_state, self.epsilon)

                # todo store the illegal / ambiguous / ... actions
                reward, done = self._init_local_train_loop()
                if self.__nb_env == 1:
                    # still the "hack" to have same interface between multi env and env...
                    # yeah it's a pain
                    act = act[0]

                temp_observation_obj, temp_reward, temp_done, info = env.step(act)

                if self.__nb_env == 1:
                    # dirty hack to wrap them into list
                    temp_observation_obj = [temp_observation_obj]
                    temp_reward = np.array([temp_reward], dtype=np.float32)
                    temp_done = np.array([temp_done], dtype=np.bool)
                    info = [info]

                new_state = self._convert_obs_train(temp_observation_obj)
                self._updage_illegal_ambiguous(training_step, info)
                done, reward, total_reward, alive_frame, epoch_num \
                    = self._update_loop(done, temp_reward, temp_done, alive_frame, total_reward, reward, epoch_num)

                # update the replay buffer
                self._store_new_state(initial_state, pm_i, reward, done, new_state)

                # now train the model
                if not self._train_model(training_step):
                    # infinite loss in this case
                    raise RuntimeError("ERROR INFINITE LOSS")

                # Save the network every 1000 iterations
                if training_step % SAVING_NUM == 0 or training_step == iterations - 1:
                    self.save(save_path)

                # save some information to tensorboard
                alive_frames[epoch_num] = np.mean(alive_frame)
                total_rewards[epoch_num] = np.mean(total_reward)
                self._store_action_played_train(training_step, pm_i)

                self._save_tensorboard(training_step, epoch_num, UPDATE_FREQ, total_rewards, alive_frames)
                training_step += 1
                pbar.update(1)

        self.save(save_path)

    # auxiliary functions
    # two below function: to train with multiple environments
    def _convert_obs_train(self, observations):
        """ create the observations that are used for training."""
        if self._obs_as_vect is None:
            size_obs = self.convert_obs(observations[0]).shape[1]
            self._obs_as_vect = np.zeros((self.__nb_env, size_obs), dtype=np.float32)

        for i, obs in enumerate(observations):
            self._obs_as_vect[i, :] = self.convert_obs(obs).reshape(-1)
        return self._obs_as_vect

    def _store_action_played(self, action_int):
        """if activated, this function will store the action taken by the agent."""
        if self.store_action:
            if action_int not in self.dict_action:
                act = self.action_space.all_actions[action_int]
                is_inj, is_volt, is_topo, is_line_status, is_redisp = act.get_types()
                is_dn = (not is_inj) and (not is_volt) and (not is_topo) and (not is_line_status) and (not is_redisp)
                self.dict_action[action_int] = [0, act, (is_inj, is_volt, is_topo, is_line_status, is_redisp, is_dn)]
            self.dict_action[action_int][0] += 1

            (is_inj, is_volt, is_topo, is_line_status, is_redisp, is_dn) = self.dict_action[action_int][2]
            if is_inj:
                self.nb_injection += 1
            if is_volt:
                self.nb_voltage += 1
            if is_topo:
                self.nb_topology += 1
            if is_line_status:
                self.nb_line += 1
            if is_redisp:
                self.nb_redispatching += 1
            if is_dn:
                self.nb_do_nothing += 1

    def _convert_all_act(self, act_as_integer):
        """this function converts the action given as a list of integer. It ouputs a list of valid grid2op Action"""
        res = []
        for act_id in act_as_integer:
            res.append(self.convert_act(act_id))
            self._store_action_played(act_id)
        return res

    def _load_action_space(self, path):
        """ load the action space in case the model is reloaded"""
        if not os.path.exists(path):
            raise RuntimeError("The model should be stored in \"{}\". But this appears to be empty".format(path))
        try:
            self.action_space.init_converter(
                all_actions=os.path.join(path, "action_space.npy".format(self.name)))
        except Exception as e:
            raise RuntimeError("Impossible to reload converter action space with error \n{}".format(e))

    # utilities for data reading
    def _set_chunk(self, env, nb):
        """
        to optimize the data reading process. See the official grid2op documentation for the effect of setting
        the chunk size for the environment.
        """
        env.set_chunk_size(int(max(100, nb)))

    def _train_model(self, training_step):
        """train the deep q networks."""
        self._training_param.tell_step(training_step)
        if training_step > max(self._training_param.min_observation, self._training_param.minibatch_size) and \
            self._training_param.do_train():
            # train the model
            s_batch, a_batch, r_batch, d_batch, s2_batch = self.replay_buffer.sample(self._training_param.minibatch_size)
            tf_writer = None
            if self.__graph_saved is False:
                tf_writer = self._tf_writer
            loss = self.deep_q.train(s_batch, a_batch, r_batch, d_batch, s2_batch,
                                     tf_writer)
            # save learning rate for later
            self._train_lr = self.deep_q._optimizer_model._decayed_lr('float32').numpy()
            self.__graph_saved = True
            if not np.all(np.isfinite(loss)):
                # if the loss is not finite i stop the learning
                return False
            self.deep_q.target_train()
            self._losses[training_step:] = np.sum(loss)
        return True

    def _updage_illegal_ambiguous(self, curr_step, info):
        """update the conunt of illegal and ambiguous actions"""
        self._illegal_actions_per_1000steps[curr_step % 1000] = np.sum([el["is_illegal"] for el in info])
        self._ambiguous_actions_per_1000steps[curr_step % 1000] = np.sum([el["is_ambiguous"] for el in info])

    def _store_action_played_train(self, training_step, action_id):
        """store which action were played, for tensorboard only."""
        which_row = training_step % 1000
        self._actions_per_1000steps[which_row, :] = 0
        self._actions_per_1000steps[which_row, action_id] += 1

    def _fast_forward_env(self, env, time=7*24*60/5):
        """use this functio to skip some time steps when environment is reset."""
        my_int = np.random.randint(0, min(time, env.chronics_handler.max_timestep()))
        env.fast_forward_chronics(my_int)

    def _reset_env_clean_state(self, env):
        """
        reset this environment to a proper state. This should rather be integrated in grid2op. And will probably
        be integrated partially starting from grid2op 1.0.0
        """
        # /!\ DO NOT ATTEMPT TO MODIFY OTHERWISE IT WILL PROBABLY CRASH /!\
        # /!\ THIS WILL BE PART OF THE ENVIRONMENT IN FUTURE GRID2OP RELEASE (>= 1.0.0) /!\
        # AND OF COURSE USING THIS METHOD DURING THE EVALUATION IS COMPLETELY FORBIDDEN
        if self.__nb_env > 1:
            return
        env.current_obs = None
        env.env_modification = None
        env._reset_maintenance()
        env._reset_redispatching()
        env._reset_vectors_and_timings()
        _backend_action = env._backend_action_class()
        _backend_action.all_changed()
        env._backend_action =_backend_action
        env.backend.apply_action(_backend_action)
        _backend_action.reset()
        *_, fail_to_start, info = env.step(env.action_space())
        if fail_to_start:
            # this is happening because not enough care has been taken to handle these problems
            # more care will be taken when this feature will be available in grid2op directly.
            raise Grid2OpException("Impossible to initialize the powergrid, the powerflow diverge at iteration 0. "
                                   "Available information are: {}".format(info))
        env._reset_vectors_and_timings()

    def _need_reset(self, env, observation_num, epoch_num, done, new_state):
        """perform the proper reset of the environment"""
        if self._training_param.step_increase_nb_iter > 0:
            self._max_iter_env(min(max(self._training_param.min_iter,
                                       self._training_param.max_iter_fun(self._total_sucesses)),
                                   self._training_param.max_iter))  # TODO
        self._curr_iter_env += 1
        if new_state is None:
            # it's the first ever loop
            obs = env.reset()
            if self.__nb_env == 1:
                # still hack to have same program interface between multi env and not multi env
                obs = [obs]
            new_state = self._convert_obs_train(obs)
        elif self.__nb_env > 1:
            # in multi env this is automatically handled
            pass
        elif done[0]:
            nb_ts_one_day = 24*60/5
            if False:
                # the 3-4 lines below allow to reuse the loaded dataset and continue further up in the
                try:
                    self._reset_env_clean_state(env)
                    # random fast forward between now and next day
                    self._fast_forward_env(env, time=nb_ts_one_day)
                except (StopIteration, Grid2OpException):
                    env.reset()
                    # random fast forward between now and next week
                    self._fast_forward_env(env, time=7*nb_ts_one_day)

            # update the number of time steps it has live
            ts_lived = observation_num - self._prev_obs_num
            if self._time_step_lived is not None:
                self._time_step_lived[self._prev_id] += ts_lived
            self._prev_obs_num = observation_num
            if self._training_param.oversampling_rate is not None:
                # proba = np.sqrt(1. / (self._time_step_lived +1))
                # # over sampling some kind of "UCB like" stuff
                # # https://banditalgs.com/2016/09/18/the-upper-confidence-bound-algorithm/

                # proba = 1. / (self._time_step_lived + 1)
                self._proba[:] = 1. / (self._time_step_lived ** self._training_param.oversampling_rate + 1)
                self._proba /= np.sum(self._proba)

            _prev_id = self._prev_id
            self._prev_id = None
            if _CACHE_AVAILABLE_DEEPQAGENT:
                if isinstance(env.chronics_handler.real_data, MultifolderWithCache):
                    self._prev_id = env.chronics_handler.real_data.sample_next_chronics(self._proba)
            if self._prev_id is None:
                self._prev_id = _prev_id + 1
                self._prev_id %= self._time_step_lived.shape[0]

            env.reset()
            if self._nb_chosen is not None:
                self._nb_chosen[self._prev_id] += 1

            # random fast forward between now and next week
            if self._training_param.random_sample_datetime_start is not None:
                self._fast_forward_env(env, time=self._training_param.random_sample_datetime_start)

            self._curr_iter_env = 0
            obs = [env.current_obs]
            new_state = self._convert_obs_train(obs)
            if epoch_num % len(env.chronics_handler.real_data.subpaths) == 0:
                # re shuffle the data
                env.chronics_handler.shuffle(lambda x: x[np.random.choice(len(x), size=len(x), replace=False)])
        return new_state

    def _init_replay_buffer(self):
        """create and initialized the replay buffer"""
        self.replay_buffer = ReplayBuffer(self._training_param.buffer_size)

    def _store_new_state(self, initial_state, predict_movement_int, reward, done, new_state):
        """store the new state in the replay buffer"""
        # vectorized version of the previous code
        for i_s, pm_i, reward, done, new_state in zip(initial_state, predict_movement_int, reward, done, new_state):
            self.replay_buffer.add(i_s,
                                   pm_i,
                                   reward,
                                   done,
                                   new_state)

    def _max_iter_env(self, new_max_iter):
        """update the number of maximum iteration allowed."""
        self._max_iter_env_ = new_max_iter

    def _next_move(self, curr_state, epsilon):
        pm_i, pq_v = self.deep_q.predict_movement(curr_state, epsilon)
        act = self._convert_all_act(pm_i)
        return pm_i, pq_v, act

    def _init_global_train_loop(self):
        alive_frame = np.zeros(self.__nb_env, dtype=np.int)
        total_reward = np.zeros(self.__nb_env, dtype=np.float32)
        return alive_frame, total_reward

    def _update_loop(self, done, temp_reward, temp_done, alive_frame, total_reward, reward, epoch_num):
        if self.__nb_env == 1:
            # force end of episode at early stage of learning
            if self._curr_iter_env >= self._max_iter_env_:
                temp_done[0] = True
                temp_reward[0] = self._max_reward
                self._total_sucesses += 1

        done = temp_done
        alive_frame[done] = 0
        total_reward[done] = 0.
        self._reset_num += np.sum(done)
        if self._reset_num >= self.__nb_env:
            # increase the "global epoch num" represented by "epoch_num" only when on average
            # all environments are "done"
            epoch_num += 1
            self._reset_num = 0

        total_reward[~done] += temp_reward[~done]
        alive_frame[~done] += 1
        return done, temp_reward, total_reward, alive_frame, epoch_num

    def _init_local_train_loop(self):
        # reward, done = np.zeros(self.nb_process), np.full(self.nb_process, fill_value=False, dtype=np.bool)
        reward = np.zeros(self.__nb_env, dtype=np.float32)
        done = np.full(self.__nb_env, fill_value=False, dtype=np.bool)
        return reward, done

    def _init_deep_q(self, training_param, env):
        """
        This function serves as initializin the neural network.
        """
        if self.deep_q is None:
            self.deep_q = self._nn_archi.make_nn(training_param)
        self.init_obs_extraction(env)

    def _save_tensorboard(self, step, epoch_num, UPDATE_FREQ, epoch_rewards, epoch_alive):
        """save all the informations needed in tensorboard."""
        if self._tf_writer is None:
            return

        # Log some useful metrics every even updates
        if step % UPDATE_FREQ == 0 and epoch_num > 0:
            if step % (10 * UPDATE_FREQ) == 0:
                # print the top k scenarios the "hardest" (ie chosen the most number of times
                if self.verbose:
                    top_k = 10
                    if self._nb_chosen is not None:
                        array_ = np.argsort(self._nb_chosen)[-top_k:][::-1]
                        print("hardest scenarios\n{}".format(array_))
                        print("They have been chosen respectively\n{}".format(self._nb_chosen[array_]))
                        # print("Associated proba are\n{}".format(self._proba[array_]))
                        print("The number of timesteps played is\n{}".format(self._time_step_lived[array_]))
                        print("avg (accross all scenarios) number of timsteps played {}"
                              "".format(np.mean(self._time_step_lived)))
                        print("Time alive: {}".format(self._time_step_lived[array_] / (self._nb_chosen[array_] + 1)))
                        print("Avg time alive: {}".format(np.mean(self._time_step_lived / (self._nb_chosen + 1 ))))

            with self._tf_writer.as_default():
                last_alive = epoch_alive[(epoch_num-1)]
                last_reward = epoch_rewards[(epoch_num-1)]

                mean_reward = np.nanmean(epoch_rewards[:epoch_num])
                mean_alive = np.nanmean(epoch_alive[:epoch_num])

                mean_reward_30 = mean_reward
                mean_alive_30 = mean_alive
                mean_reward_100 = mean_reward
                mean_alive_100 = mean_alive

                tmp = self._actions_per_1000steps > 0
                tmp = tmp.sum(axis=0)
                nb_action_taken_last_1000_step = np.sum(tmp > 0)

                nb_illegal_act = np.sum(self._illegal_actions_per_1000steps)
                nb_ambiguous_act = np.sum(self._ambiguous_actions_per_1000steps)

                if epoch_num >= 100:
                    mean_reward_100 = np.nanmean(epoch_rewards[(epoch_num-100):epoch_num])
                    mean_alive_100 = np.nanmean(epoch_alive[(epoch_num-100):epoch_num])

                if epoch_num >= 30:
                    mean_reward_30 = np.nanmean(epoch_rewards[(epoch_num-30):epoch_num])
                    mean_alive_30 = np.nanmean(epoch_alive[(epoch_num-30):epoch_num])

                # to ensure "fair" comparison between single env and multi env
                step_tb = step  # * self.__nb_env
                # if multiply by the number of env we have "trouble" with random exploration at the beginning
                # because it lasts the same number of "real" steps

                # show first the Mean reward and mine time alive (hence the upper case)
                tf.summary.scalar("Mean_alive_30", mean_alive_30, step_tb,
                                  description="Mean reward over the last 30 epochs")
                tf.summary.scalar("Mean_reward_30", mean_reward_30, step_tb,
                                  description="Mean number of timesteps sucessfully manage the last 30 epochs")

                # then it's alpha numerical order, hence the "z_" in front of some information
                tf.summary.scalar("loss", self._losses[step], step_tb,
                                  description="last training loss")

                tf.summary.scalar("last_alive", last_alive, step_tb,
                                  description="last number of timestep during which the agent stayed alive")
                tf.summary.scalar("last_reward", last_reward, step_tb,
                                  description="last reward get by the agent")

                tf.summary.scalar("mean_reward", mean_reward, step_tb)
                tf.summary.scalar("mean_alive", mean_alive, step_tb)

                tf.summary.scalar("mean_reward_100", mean_reward_100, step_tb,
                                  description="Mean reward over the last 100 epochs")
                tf.summary.scalar("mean_alive_100", mean_alive_100, step_tb,
                                  description="Mean number of timesteps sucessfully manage the last 100 epochs")

                tf.summary.scalar("nb_differentaction_taken_1000", nb_action_taken_last_1000_step, step_tb,
                                  description="Number of different actions played the past 1000 steps")
                tf.summary.scalar("nb_illegal_act", nb_illegal_act, step_tb,
                                  description="Number of illegal actions played the past 1000 steps")
                tf.summary.scalar("nb_ambiguous_act", nb_ambiguous_act, step_tb,
                                  description="Number of ambiguous actions played the past 1000 steps")
                tf.summary.scalar("nb_total_success", self._total_sucesses, step_tb,
                                  description="Number of times I reach the end of scenario (no game over)")

                tf.summary.scalar("z_lr", self._train_lr, step_tb,
                                  description="current learning rate")
                tf.summary.scalar("z_epsilon", self.epsilon, step_tb,
                                  description="current epsilon (of the epsilon greedy)")
                tf.summary.scalar("z_max_iter", self._max_iter_env_, step_tb,
                                  description="maximum number of time steps before deciding a scenario is over (=win)")
                tf.summary.scalar("z_total_episode", epoch_num, step_tb,
                                  description="total number of episode played (~number of \"reset\")")

                if self.store_action:
                    nb_ = 10  # reset the frequencies every nb_ saving
                    self._nb_updated_act_tensorboard += UPDATE_FREQ
                    tf.summary.scalar("zz_freq_inj", self.nb_injection / self._nb_updated_act_tensorboard, step_tb)
                    tf.summary.scalar("zz_freq_voltage", self.nb_voltage / self._nb_updated_act_tensorboard, step_tb)
                    tf.summary.scalar("z_freq_topo", self.nb_topology / self._nb_updated_act_tensorboard, step_tb)
                    tf.summary.scalar("z_freq_line_status", self.nb_line / self._nb_updated_act_tensorboard, step_tb)
                    tf.summary.scalar("z_freq_redisp", self.nb_redispatching / self._nb_updated_act_tensorboard, step_tb)
                    tf.summary.scalar("z_freq_do_nothing", self.nb_do_nothing / self._nb_updated_act_tensorboard, step_tb)
                    if step % (nb_ * UPDATE_FREQ) == 0:
                        self.nb_injection = 0
                        self.nb_voltage = 0
                        self.nb_topology = 0
                        self.nb_line = 0
                        self.nb_redispatching = 0
                        self.nb_do_nothing = 0
                        self._nb_updated_act_tensorboard = 0

                if self._time_step_lived is not None:
                    tf.summary.histogram(
                        "timestep_lived", self._time_step_lived, step=step_tb, buckets=None,
                        description="number of time steps lived for all scenarios"
                    )
                if self._nb_chosen is not None:
                    tf.summary.histogram(
                        "nb_chosen", self._nb_chosen, step=step_tb, buckets=None,
                        description="number of times this scenarios has been played"
                    )



__all__ = [
    "DuelQLeapNet",
    "evaluate",
    "train",
    "DuelQLeapNet_NN"
]

from l2rpn_baselines.DuelQLeapNet.DuelQLeapNet import DuelQLeapNet
from l2rpn_baselines.DuelQLeapNet.evaluate import evaluate
from l2rpn_baselines.DuelQLeapNet.train import train
from l2rpn_baselines.DuelQLeapNet.DuelQLeapNet_NN import DuelQLeapNet_NN



import warnings

__all__ = [
    "AsynchronousActorCritic",
    "evaluate",
    "train",
    "Action_reduced_list",
    "Runner",
    "run_grid2viz",
    "user_environment_make"
]

from l2rpn_baselines.AsynchronousActorCritic.AsynchronousActorCritic import *
from l2rpn_baselines.AsynchronousActorCritic.evaluate import evaluate

try:
    from l2rpn_baselines.AsynchronousActorCritic.train import train
    __all__.append("train")
except ImportError as exc_:
    warnings.warn("AsynchronousActorCritic: impossible to load the \"train\" function because of missing dependencies. The error was: \n{}".format(exc_))

try:
    from l2rpn_baselines.AsynchronousActorCritic.Runner import Runner
    __all__.append("Runner")
except ImportError as exc_:
    warnings.warn("AsynchronousActorCritic: impossible to load the \"Runner\" function because of missing dependencies. The error was: \n{}".format(exc_))



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

# tf2.0 friendly
import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=FutureWarning)
    from tensorflow.keras.models import Sequential, Model
    from tensorflow.keras.layers import Activation, Dense
    from tensorflow.keras.layers import Input

from l2rpn_baselines.utils import BaseDeepQ, TrainingParam


class DeepQ_NN(BaseDeepQ):
    """
    Constructs the desired deep q learning network

    Attributes
    ----------
    schedule_lr_model:
        The schedule for the learning rate.
    """

    def __init__(self,
                 nn_params,
                 training_param=None):
        if training_param is None:
            training_param = TrainingParam()
        BaseDeepQ.__init__(self,
                           nn_params,
                           training_param)
        self.schedule_lr_model = None
        self.construct_q_network()

    def construct_q_network(self):
        """
        The network architecture can be changed with the :attr:`l2rpn_baselines.BaseDeepQ.nn_archi`

        This function will make 2 identical models, one will serve as a target model, the other one will be trained
        regurlarly.
        """
        self._model = Sequential()
        input_layer = Input(shape=(self._nn_archi.observation_size,),
                            name="state")
        lay = input_layer
        for lay_num, (size, act) in enumerate(zip(self._nn_archi.sizes, self._nn_archi.activs)):
            lay = Dense(size, name="layer_{}".format(lay_num))(lay)  # put at self.action_size
            lay = Activation(act)(lay)

        output = Dense(self._action_size, name="output")(lay)

        self._model = Model(inputs=[input_layer], outputs=[output])
        self._schedule_lr_model, self._optimizer_model = self.make_optimiser()
        self._model.compile(loss='mse', optimizer=self._optimizer_model)

        self._target_model = Model(inputs=[input_layer], outputs=[output])
        self._target_model.set_weights(self._model.get_weights())


__all__ = [
    "Template",
    "DoubleDuelingDQN",
    "DoubleDuelingRDQN",
    "DoNothing",
    "SliceRDQN",
    "DeepQSimple",
    "DuelQSimple",
    "SAC",
    # contribution
    "PandapowerOPFAgent",
    "Geirina",
    "AsynchronousActorCritic",
    # utilitary scripts
    "utils"
]

__version__ = "0.4.0"
